import asyncio
import base64
import json
from io import BytesIO
from typing import Optional

import aiohttp
from PIL import Image

from .skin import Skin
from .utils import (
    fetch_optifine_cape,
    fetch_labymod_cape,
    fetch_5zig_cape,
    fetch_minecraftcapes_cape,
    fetch_tlauncher_cape,
    name_to_uuid,
    uuid_to_undashed,
)
from .errors import InvalidPlayer


class Player:
    """Class representing a minecraft player
    This has to be created before a skin can be rendered

    Parameters
    ----------
    uuid: str
        UUID of the player (Not needed if name is given
    name: str
        Username of the player (Not needed if UUID is given)
    raw_skin: PIL.Image.Image
        Raw skin image of the player (64x64px)
    raw_cape: PIL.Image.Image
        Raw cape image of the player (64x32px)
    session: aiohttp.ClientSession
        ClientSession to use for requests
    """
    def __init__(
            self,
            uuid: str = None,
            name: str = None,
            raw_skin: Image.Image = None,
            raw_cape: Image.Image = None,
            session: aiohttp.ClientSession = None
    ):
        if uuid is None and name is None:
            raise ValueError("Pass a username or UUID")

        self._uuid: Optional[str] = uuid
        self._username: Optional[str] = name

        self._skin: Optional[Skin] = None
        self._raw_skin: Optional[Image.Image] = raw_skin
        self._raw_capes: dict = {
            "default": raw_cape,
            "mojang": None,
            "optifine": None,
            "labymod": None,
            "5zig": None,
            "minecraftcapes": None,
            "tlauncher": None,
        }

        self._session: Optional[aiohttp.ClientSession] = session
        self._close_session: bool = False

        self._ready: asyncio.Event = asyncio.Event()
        if self._uuid:
            self._uuid = uuid_to_undashed(self._uuid)

            if len(self._uuid) != 32:
                raise ValueError("UUID seems to be invalid.")

    def __repr__(self):
        return f"<Player (UUID={self.uuid}) (name={self.name}) (skin={self._skin})>"

    @property
    def uuid(self):
        """The players UUID\n
        Either passed when creating this class or fetched from the MojangAPI"""
        return self._uuid

    @property
    def name(self):
        """The players name"""
        return self._username

    @property
    def skin(self):
        """The players :class:`Skin`"""
        return self._skin

    @property
    def capes(self):
        """A dict representing this player's capes"""
        return self._raw_capes

    @property
    def mojang_cape(self):
        """The player's mojang cape"""
        return self._raw_capes["mojang"]

    @property
    def optifine_cape(self):
        """The player's optifine cape"""
        return self._raw_capes["optifine"]

    @property
    def labymod_cape(self):
        """The player's labymod cape"""
        return self._raw_capes["labymod"]

    @property
    def minecraftcapes_cape(self):
        """The player's MinecraftCapes cape"""
        return self._raw_capes["minecraftcapes"]

    @property
    def zig_cape(self):
        """The player's 5Zig cape"""
        return self._raw_capes["5zig"]

    @property
    def tlauncher_cape(self):
        """The player's TLauncher cape"""
        return self._raw_capes["tlauncher"]

    def set_skin(self, skin: "Skin"):
        """Manually overwrite/set this players skin

        Parameters
        ----------
        skin: Skin
            The new skin
        """
        self._skin = skin

    async def initialize(self):
        """Initializes the player class

        This function is an initializer which helps to get various details about the player with just one method.
        Once this function has finished running, the corresponding :py:class:`Player` object is
        guaranteed to have a name, UUID and skin (includes the cape if available) associated to it
        (if the given username and/or UUID is valid of course).

        Warning
        -------
        This function does one to four API calls to the mojang API:
            -> (1.) Obtain the players UUID by name (Only if no UUID is given)\n
            -> 2. Get the players profile\n
            -> (3.) Get the players skin (Only if no skin is given)\n
            -> (4.) Get the players cape (Only if the player actually has a cape)\n
        Rate limits of the API are unknown but expected to be somewhere close to 6000 requests per 10 minutes.

        Raises
        ------
        errors.InvalidPlayer
            Player does not seem to be valid
        """
        if not self._session:
            self._session = aiohttp.ClientSession()
            self._close_session = True

        if self._uuid is None:
            uuid = await name_to_uuid(self._username, self._session)
            if uuid:
                self._uuid = uuid_to_undashed(uuid)

        textures = None
        if self._uuid is not None and (self._raw_skin is None or self._raw_capes["default"] is None):
            async with self._session.get(
                f"https://sessionserver.mojang.com/session/minecraft/profile/{self._uuid}"
            ) as resp:
                if resp.status == 200:
                    resp_dict = await resp.json()

                    self._username = resp_dict["name"] if self._username is None else self._username

                    for p in resp_dict["properties"]:
                        if p["name"] == "textures":
                            textures = json.loads(base64.b64decode(p["value"]))["textures"]
                            break

                    else:
                        raise InvalidPlayer()

        if textures is not None:
            _raw_skin_url = textures["SKIN"]["url"]
            _raw_cape_url = textures["CAPE"]["url"] if "CAPE" in textures.keys() else None

            if not self._raw_skin:
                async with self._session.get(_raw_skin_url) as resp:
                    self._raw_skin = Image.open(BytesIO(await resp.read()))

            if not self._raw_capes["default"]:
                if _raw_cape_url is not None:
                    async with self._session.get(_raw_cape_url) as resp:
                        self._raw_capes["default"] = Image.open(BytesIO(await resp.read()))
                        self._raw_capes["mojang"] = self._raw_capes["default"]
                else:
                    self._raw_capes["default"] = None

            self._skin = Skin(
                raw_skin=self._raw_skin,
                raw_skin_url=_raw_skin_url,
                raw_cape=self._raw_capes["default"],
                raw_cape_url=_raw_cape_url
            )

        if self._close_session:
            await self._session.close()

        self._ready.set()

    async def wait_for_fully_constructed(self):
        """Returns as soon as the initialize function finished

        Waiting for this guarantees that the skin has been fetched from the api"""
        try:
            await asyncio.wait_for(self._ready.wait(), timeout=60)
        except asyncio.TimeoutError:
            raise asyncio.TimeoutError

    async def fetch_optifine_cape(self):
        """Fetches the players optifine cape and stores it to :py:attr:`Player.optifine_cape`
        
        This is basically just an alias for :py:func:`utils.fetch_optifine_cape`"""
        cape = await fetch_optifine_cape(self)
        if cape is not None:
            self._raw_capes["optifine"] = cape

    async def fetch_labymod_cape(self):
        """Fetches the players labymod cape and stores it to :py:attr:`Player.labymod_cape`

        This is basically just an alias for :py:func:`utils.fetch_labymod_cape`"""
        cape = await fetch_labymod_cape(self)
        if cape is not None:
            self._raw_capes["labymod"] = cape

    async def fetch_minecraftcapes_cape(self):
        """Fetches the players MinecraftCapes cape and stores it to :py:attr:`Player.minecraftcapes_cape`

        This is basically just an alias for :py:func:`utils.fetch_minecraftcapes_cape`"""
        cape = await fetch_minecraftcapes_cape(self)
        if cape is not None:
            self._raw_capes["minecraftcapes"] = cape

    async def fetch_5zig_cape(self):
        """Fetches the players 5Zig cape and stores it to :py:attr:`Player.zig_cape`

        This is basically just an alias for :py:func:`utils.fetch_5zig_cape`"""
        cape = await fetch_5zig_cape(self)
        if cape is not None:
            self._raw_capes["5zig"] = cape

    async def fetch_tlauncher_cape(self):
        """Fetches the players TLauncher cape and stores it to :py:attr:`Player.tlauncher_cape`

        This is basically just an alias for :py:func:`utils.fetch_tlauncher_cape`"""
        cape = await fetch_tlauncher_cape()
        if cape is not None:
            self._raw_capes["tlauncher"] = cape



import base64

from typing import Optional
from PIL import Image, ImageOps
from io import BytesIO

from .skin_render import Render
from .errors import NoRenderedSkin

class Skin:
    """
    Tip
    ----
    It's best practice to save the players skin to a database using :py:func:`encodeb64` and then
    reinitialize this class using :py:func:`decodeb64`. This way no API call is needed.\n
    Alternatively you can manually instantiate by passing a raw skin image.

    Parameters
    ----------
    raw_skin: PIL.Image.Image
        The raw skin image
    raw_cape: PIL.Image.Image
        The raw cape image
    """

    def __init__(
            self,
            raw_skin,
            raw_cape=None,
            raw_skin_url=None,
            raw_cape_url=None,
            name=None,
    ):
        self._raw_skin: Image.Image = raw_skin
        self._raw_skin_url: Optional[str] = raw_skin_url
        self._raw_cape: Optional[Image.Image] = None
        self._raw_cape_url: Optional[str] = raw_cape_url
        self._name: Optional[str] = name

        self._skin: Optional[Image.Image] = None
        self._head: Optional[Image.Image] = None

        if raw_cape is not None:
            self.set_cape(raw_cape)

        if self._raw_skin.mode != "RGBA":  # Converting skins to RGBA
            self._raw_skin = self._raw_skin.convert(mode="RGBA")

        if self._raw_skin.height == 32:  # old skin format
            new_skin_im = Image.new("RGBA", (64, 64), (0, 0, 0, 0))
            new_skin_im.paste(self._raw_skin, (0, 0))

            for i in range(2):
                f1 = 40 if i != 0 else 0
                f2 = 16 if i != 0 else 0

                new_skin_im.paste(ImageOps.mirror(self._raw_skin.crop((4+f1, 16, 8+f1, 20))), (20+f2, 48))
                new_skin_im.paste(ImageOps.mirror(self._raw_skin.crop((8+f1, 16, 12+f1, 20))), (24+f2, 48))
                new_skin_im.paste(ImageOps.mirror(self._raw_skin.crop((8+f1, 20, 12+f1, 32))), (16+f2, 52))
                new_skin_im.paste(ImageOps.mirror(self._raw_skin.crop((12+f1, 20, 16+f1, 32))), (20+f2, 52))
                new_skin_im.paste(ImageOps.mirror(self._raw_skin.crop((4+f1, 20, 8+f1, 32))), (24+f2, 52))
                new_skin_im.paste(ImageOps.mirror(self._raw_skin.crop((0+f1, 20, 4+f1, 32))), (28+f2, 52))

            self._raw_skin = new_skin_im

    def __repr__(self):
        return f"<Skin (slim={self.is_slim}) (has_cape={self.has_cape})>"

    @property
    def skin(self):
        """The last skin which has been rendered using :py:func:`render_skin`"""
        return self._skin

    @property
    def head(self):
        """The last head which has been rendered using :py:func:`render_head`"""
        return self._head

    @property
    def raw_skin(self):
        """The players raw skin image"""
        return self._raw_skin

    @property
    def raw_skin_url(self):
        """The players skins url. Returns None if the skin has been manually passed"""
        return self._raw_skin_url

    @property
    def raw_cape(self):
        """The players raw cape image"""
        return self._raw_cape

    @property
    def raw_cape_url(self):
        """The players capes url. Returns None if the player doesn't have a cape or the cape has been manually passed"""
        return self._raw_cape_url

    @property
    def has_cape(self):
        """Whether the player has a cape"""
        return bool(self._raw_cape)

    @property
    def is_slim(self):
        """Whether the skin is slim (Alex type) or classic (Steve type)

        Only difference being the width of the arms (3px - 4px)"""
        return not bool(self._raw_skin.getpixel((46, 52))[3])

    def set_cape(self, cape: Image.Image):
        """Change the players cape

        Parameters
        ----------
        cape: PIL.Image.Image
            The new cape image (64x32px)
        """
        # attempt to detect wrongly scaled mojang capes
        if cape.width / cape.height == 2 and (cape.width != 64 or cape.height != 32):
            cape.resize((64, 32), resample=Image.LANCZOS)

        if cape.width == 22 and cape.height == 17:  # Labymod
            pass

        if cape.width == 64 and cape.height > 32:  # MinecraftCapes animated
            pass

        if cape.mode != "RGBA":  # Converting capes to RGBA
            cape = cape.convert(mode="RGBA")

        self._raw_cape = cape

    def show(self):
        """Shows the last rendered skin

        Alias for :py:func:`Skin.skin.show()`

        Raises
        ------
        NoRenderedSkin
            No skin present. Generate a render using :py:func:`render_skin`
        """
        if self._skin:
            self._skin.show()
        else:
            raise NoRenderedSkin()

    def encodeb64(self):
        """Base64 encodes the players skin and cape

        This allows for better caching and persistent storage in a database
        A :class:`Skin` class can then be recreated using :py:func:`decodeb64`

        Returns
        -------
        str
            The players skin and cape in format {raw_skin}-{raw_cape}"""
        with BytesIO() as buffered:
            self._raw_skin.save(buffered, format="PNG")
            buffered.seek(0)
            im_skin = buffered.getvalue()
        if self._raw_cape:
            with BytesIO() as buffered:
                self._raw_cape.save(buffered, format="PNG")
                buffered.seek(0)
                im_cape = buffered.getvalue()
        else:
            im_cape = None

        bytelist = [base64.b64encode(im_skin), base64.b64encode(im_cape)] if im_cape else [base64.b64encode(im_skin)]
        return b';'.join(bytelist).decode()

    @classmethod
    def decodeb64(cls, b64: str):
        """Create an instance of this class from a saved b64 string

        Parameters
        ----------
        b64: str
            The base64 encoded raw_skin image or raw_skin and raw_cape separated with a ";".
            Can be obtained from :py:func:`encodeb64`

        Returns
        -------
        :class:`Skin`
        """
        skin, cape = b64.split(";")
        skin.encode()
        if cape:
            cape.encode()

        im_str = base64.b64decode(skin)
        buffered = BytesIO(im_str)
        im_skin = Image.open(buffered)

        if cape:
            im_str = base64.b64decode(cape)
            buffered = BytesIO(im_str)
            im_cape = Image.open(buffered)
        else:
            im_cape = None

        return cls(raw_skin=im_skin, raw_cape=im_cape)

    async def render_skin(
            self,
            vr: int = 25,
            hr: int = 35,
            hrh: int = 0,
            vrll: int = 0,
            vrrl: int = 0,
            vrla: int = 0,
            vrra: int = 0,
            vrc: int = 30,
            ratio: int = 12,
            display_hair: bool = True,
            display_second_layer: bool = True,
            display_cape: bool = True,
            aa: bool = False,
    ) -> Optional[Image.Image]:
        """Render a full body skin

        Parameters
        ----------
        vr: int
            Vertical rotation of the output image
        hr: int
            Horizontal rotation of the output image
        hrh: int
            Horizontal head rotation
        vrll: int
            Vertical rotation of the left leg
        vrrl: int
            Vertical rotation of the right leg
        vrla: int
            Vertical rotation of the left arm
        vrra: int
            Vertical rotation of the right arm
        vrc: int
            Vertical rotation of the cape
            Not actually in degrees, use random values please until you find one you like
        ratio: int
            Resolution of the returned image
        display_hair: bool
            Whether the second head layer is displayed
        display_second_layer: bool
            Whether the second skin layer is displayed
        display_cape: bool
            Whether the player's cape is shown
        aa: bool
            Antialiasing: smoothens the corners a bit

        Returns
        -------
        PIL.Image.Image
            The rendered skin
        """
        render = Render(
            player=self,
            vr=vr,
            hr=hr,
            hrh=hrh,
            vrll=vrll,
            vrrl=vrrl,
            vrla=vrla,
            vrra=vrra,
            vrc=vrc,
            ratio=ratio,
            head_only=False,
            display_hair=display_hair,
            display_layers=display_second_layer,
            display_cape=display_cape,
            aa=aa,
        )
        im = await render.get_render()
        self._skin = im
        return im

    async def render_head(
            self,
            vr: int = 25,
            hr: int = 35,
            ratio: int = 12,
            display_hair: bool = True,
            aa: bool = False,
    ) -> Optional[Image.Image]:
        """Render the players head

        Parameters
        ----------
        vr: int
            Vertical rotation of the output image
        hr: int
            Horizontal rotation of the output image
        ratio: int
            Resolution of the returned image
        display_hair: bool
            Whether or not the second head layer should be displayed
        aa: bool
            Antialiasing: smoothens the corners a bit

        Returns
        -------
        PIL.Image.Image
            The rendered head
        """
        render = Render(
            player=self,
            vr=vr,
            hr=hr,
            ratio=ratio,
            head_only=True,
            display_hair=display_hair,
            aa=aa,
        )
        im = await render.get_render()
        self._head = im
        return im



from .player import Player
from .skin import Skin

from .utils import (
    uuid_to_dashed,
    uuid_to_undashed,
    name_to_uuid,
    uuid_to_name,
    fetch_skin,
    fetch_optifine_cape,
    fetch_labymod_cape,
    fetch_5zig_cape,
    fetch_minecraftcapes_cape,
    get_players_by_name,
)



from distutils.core import setup

setup(
    name="MinePI",
    packages=["minepi"],
    version="0.5.0",
    license="MIT",
    description="Minecraft utility library.",
    author="benno1237, honigkuchen",
    author_email="benno.kollreider@gmail.com",
    url="https://github.com/benno1237/MinePI",
    download_url="https://github.com/benno1237/MinePI/archive/refs/tags/0.5.0.tar.gz",
    keywords=["Minecraft", "Skin", "Render", "Head", "UUID"],
    install_requires=[
        "aiohttp",
        "Pillow",
    ],
    extras_require={
        "docs": [
            "sphinx",
            "sphinx-rtd-theme",
            "sphinx-prompt",
            "autodocsumm"
        ]
    },
    classifiers=[
    'Development Status :: 3 - Alpha',     
    'Intended Audience :: Developers',      
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',  
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    ],
)



from math import radians, sin, cos
from PIL import Image, ImageDraw
import asyncio
import typing

if typing.TYPE_CHECKING:
    from . import Skin

def is_not_existing(dic, key1=None, key2=None, key3=None):
    try:
        if key1 is None:
            dic
        if key2 is None:
            dic[key1]
        elif key3 is None:
            dic[key1][key2]
        else:
            dic[key1][key2][key3]
        return False
    except KeyError:
        return True


def append_dict(dic, key1, key2, key3, value):
    if is_not_existing(dic, key1, key2, key3):
        try:
            dic[key1][key2][key3] = value
        except KeyError:
            try:
                dic[key1][key2] = {}
                dic[key1][key2][key3] = value
            except KeyError:
                dic[key1] = {}
                dic[key1][key2] = {}
                dic[key1][key2][key3] = value
    return dic


class Render:
    def __init__(
            self,
            player: "Skin",
            vr: int = 0,
            hr: int = 0,
            hrh: int = 0,
            vrll: int = 0,
            vrrl: int = 0,
            vrla: int = 0,
            vrra: int = 0,
            vrc: int = 30,
            ratio: int = 12,
            head_only: bool = False,
            display_hair: bool = False,
            display_layers: bool = False,
            display_cape: bool = False,
            aa: bool = False,
    ):
        self.vr = vr
        self.hr = hr
        self.hrh = hrh
        self.vrll = vrll
        self.vrrl = vrrl
        self.vrla = vrla
        self.vrra = vrra
        self.vrc = vrc
        self.head_only = head_only
        self.ratio = ratio
        self.display_hair = display_hair
        self.display_cape = display_cape if player.raw_cape else False
        self.layers = display_layers
        self.player = player
        self.aa = aa
        self.rendered_image = None

        self.loop = asyncio.get_event_loop()
        self.cube_points = []
        self.polygons = {}

        self.cos_a = None
        self.sin_a = None
        self.cos_b = None
        self.sin_b = None

        self.min_x = 0
        self.max_x = 0
        self.min_y = 0
        self.max_y = 0

    async def get_render(self):
        skin = self.player.raw_skin
        hd_ratio = int(skin.size[0] / 64)

        def render_skin(skin):
            self.calculate_angles()
            self.determine_faces()
            self.generate_polygons(hd_ratio, skin, self.player.raw_cape)
            self.member_rotation(hd_ratio)
            self.create_project_plan()

            im = self.display_image()
            return im

        im = await self.loop.run_in_executor(
            None,
            render_skin,
            skin
        )

        return im

    def calculate_angles(self):
        self.body_angles = {}
        alpha = radians(self.vr)
        beta = radians(self.hr)

        self.cos_a, self.sin_a = cos(alpha), sin(alpha)
        self.cos_b, self.sin_b = cos(beta), sin(beta)

        self.body_angles["torso"] = (cos(0), sin(0), cos(0), sin(0))
        self.body_angles["torso_layer"] = (cos(0), sin(0), cos(0), sin(0))

        alpha_cape = (-radians(self.vrc))
        self.body_angles["cape"] = (cos(alpha_cape), sin(alpha_cape), cos(0), sin(0))
        #self.body_angles["cape"] = (cos(0), sin(alpha_cape), cos(0), sin(0))

        alpha_head = 0
        beta_head = radians(self.hrh)
        self.body_angles["head"] = (cos(alpha_head), sin(alpha_head), cos(beta_head), sin(beta_head))
        self.body_angles["helmet"] = (cos(alpha_head), sin(alpha_head), cos(beta_head), sin(beta_head))

        alpha_r_arm = radians(self.vrra)
        beta_r_arm = 0
        self.body_angles["r_arm"] = (cos(alpha_r_arm), sin(alpha_r_arm), cos(beta_r_arm), sin(beta_r_arm))
        self.body_angles["r_arm_layer"] = (cos(alpha_r_arm), sin(alpha_r_arm), cos(beta_r_arm), sin(beta_r_arm))

        alpha_l_arm = radians(self.vrla)
        beta_l_arm = 0
        self.body_angles["l_arm"] = (cos(alpha_l_arm), sin(alpha_l_arm), cos(beta_l_arm), sin(beta_l_arm))
        self.body_angles["l_arm_layer"] = (cos(alpha_l_arm), sin(alpha_l_arm), cos(beta_l_arm), sin(beta_l_arm))

        alpha_r_leg = radians(self.vrrl)
        beta_r_leg = 0
        self.body_angles["r_leg"] = (cos(alpha_r_leg), sin(alpha_r_leg), cos(beta_r_leg), sin(beta_r_leg))
        self.body_angles["r_leg_layer"] = (cos(alpha_r_leg), sin(alpha_r_leg), cos(beta_r_leg), sin(beta_r_leg))

        alpha_l_leg = radians(self.vrll)
        beta_l_leg = 0
        self.body_angles["l_leg"] = (cos(alpha_l_leg), sin(alpha_l_leg), cos(beta_l_leg), sin(beta_l_leg))
        self.body_angles["l_leg_layer"] = (cos(alpha_l_leg), sin(alpha_l_leg), cos(beta_l_leg), sin(beta_l_leg))

    def determine_faces(self):
        self.visible_faces = {
            "head": {"front": [], "back": {}},
            "torso": {"front": [], "back": {}},
            "torso_layer": {"front": [], "back": {}},
            "r_arm": {"front": [], "back": {}},
            "r_arm_layer": {"front": [], "back": {}},
            "l_arm": {"front": [], "back": {}},
            "l_arm_layer": {"front": [], "back": {}},
            "r_leg": {"front": [], "back": {}},
            "r_leg_layer": {"front": [], "back": {}},
            "l_leg": {"front": [], "back": {}},
            "l_leg_layer": {"front": [], "back": {}},
            "cape": {"front": [], "back": {}}
        }

        all_faces = ["back", "right", "top", "front", "left", "bottom"]

        for k, v in self.visible_faces.items():
            cube_max_depth_faces = None
            self.set_cube_points()

            for cube_point in self.cube_points:
                cube_point[0].pre_project(0, 0, 0, self.body_angles[k][0], self.body_angles[k][1],
                                          self.body_angles[k][2], self.body_angles[k][3])
                cube_point[0].project()

                if (not cube_max_depth_faces) or (cube_max_depth_faces[0].get_depth() > cube_point[0].get_depth()):
                    cube_max_depth_faces = cube_point

            v["back"] = cube_max_depth_faces[1]
            v["front"] = [face for face in all_faces if face not in v["back"]]

        self.front_faces = self.visible_faces["torso"]["front"]
        self.back_faces = [face for face in all_faces if face not in self.front_faces]

    def set_cube_points(self):
        self.cube_points.append(
            (
                Point(
                    self,
                    {
                        "x": 0,
                        "y": 0,
                        "z": 0
                    }
                ),
                ["back", "right", "top"]
            )
        )

        self.cube_points.append(
            (
                Point(
                    self,
                    {
                        "x": 0,
                        "y": 0,
                        "z": 1
                    }
                ),
                ["front", "right", "top"]
            )
        )

        self.cube_points.append(
            (
                Point(
                    self,
                    {
                        "x": 0,
                        "y": 1,
                        "z": 0
                    }
                ),
                ["back", "right", "bottom"]
            )
        )

        self.cube_points.append(
            (
                Point(
                    self,
                    {
                        "x": 0,
                        "y": 1,
                        "z": 1
                    }
                ),
                ["front", "right", "bottom"]
            )
        )

        self.cube_points.append(
            (
                Point(
                    self,
                    {
                        "x": 1,
                        "y": 0,
                        "z": 0
                    }
                ),
                ["back", "left", "top"]
            )
        )

        self.cube_points.append(
            (
                Point(
                    self,
                    {
                        "x": 1,
                        "y": 0,
                        "z": 1
                    }
                ),
                ["front", "left", "top"]
            )
        )

        self.cube_points.append(
            (
                Point(
                    self,
                    {
                        "x": 1,
                        "y": 1,
                        "z": 0
                    }
                ),
                ["back", "left", "bottom"]
            )
        )

        self.cube_points.append(
            (
                Point(
                    self,
                    {
                        "x": 1,
                        "y": 1,
                        "z": 1
                    }
                ),
                ["front", "left", "bottom"]
            )
        )

    def generate_polygons(self, hd_ratio, skin, im_cape):
        self.polygons = {
            "helmet": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "head": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "torso": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "torso_layer": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "r_arm": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "r_arm_layer": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "l_arm": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "l_arm_layer": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "r_leg": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "r_leg_layer": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "l_leg": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "l_leg_layer": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []},
            "cape": {"front": [], "back": [], "top": [], "bottom": [], "right": [], "left": []}
        }

        """Head"""
        volume_points = {}
        for i in range(0, 9 * hd_ratio):
            for j in range(0, 9 * hd_ratio):
                volume_points = append_dict(volume_points, i, j, -2 * hd_ratio,
                                            Point(self, {"x": i, "y": j, "z": -2 * hd_ratio}))
                volume_points = append_dict(volume_points, i, j, 6 * hd_ratio,
                                            Point(self, {"x": i, "y": j, "z": 6 * hd_ratio}))

        for j in range(0, 9 * hd_ratio):
            for k in range(-2 * hd_ratio, 7 * hd_ratio):
                volume_points = append_dict(volume_points, 0, j, k, Point(self, {"x": 0, "y": j, "z": k}))
                volume_points = append_dict(volume_points, 8 * hd_ratio, j, k,
                                            Point(self, {"x": 8 * hd_ratio, "y": j, "z": k}))

        for i in range(0, 9 * hd_ratio):
            for k in range(-2 * hd_ratio, 7 * hd_ratio):
                volume_points = append_dict(volume_points, i, 0, k, Point(self, {"x": i, "y": 0, "z": k}))
                volume_points = append_dict(volume_points, i, 8 * hd_ratio, k,
                                            Point(self, {"x": i, "y": 8 * hd_ratio, "z": k}))

        if "back" in self.visible_faces["head"]["front"]:
            for i in range(0, 8 * hd_ratio):
                for j in range(0, 8 * hd_ratio):
                    color = skin.getpixel((32 * hd_ratio - 1 - i, 8 * hd_ratio + j))
                    if color[3] != 0:
                        self.polygons["head"]["back"].append(Polygon([
                            volume_points[i][j][-2 * hd_ratio],
                            volume_points[i + 1][j][-2 * hd_ratio],
                            volume_points[i + 1][j + 1][-2 * hd_ratio],
                            volume_points[i][j + 1][-2 * hd_ratio]],
                            color))

        if "front" in self.visible_faces["head"]["front"]:
            for i in range(0, 8 * hd_ratio):
                for j in range(0, 8 * hd_ratio):
                    color = skin.getpixel((8 * hd_ratio + i, 8 * hd_ratio + j))
                    if color[3] != 0:
                        self.polygons["head"]["front"].append(Polygon([
                            volume_points[i][j][6 * hd_ratio],
                            volume_points[i + 1][j][6 * hd_ratio],
                            volume_points[i + 1][j + 1][6 * hd_ratio],
                            volume_points[i][j + 1][6 * hd_ratio]],
                            color))

        if "right" in self.visible_faces["head"]["front"]:
            for j in range(0, 8 * hd_ratio):
                for k in range(-2 * hd_ratio, 6 * hd_ratio):
                    color = skin.getpixel((k + 2 * hd_ratio, 8 * hd_ratio + j))
                    if color[3] != 0:
                        self.polygons["head"]["right"].append(Polygon([
                            volume_points[0][j][k],
                            volume_points[0][j][k + 1],
                            volume_points[0][j + 1][k + 1],
                            volume_points[0][j + 1][k]],
                            color))

        if "left" in self.visible_faces["head"]["front"]:
            for j in range(0, 8 * hd_ratio):
                for k in range(-2 * hd_ratio, 6 * hd_ratio):
                    color = skin.getpixel(((24 * hd_ratio - 1) - k - 2 * hd_ratio, 8 * hd_ratio + j))
                    if color[3] != 0:
                        self.polygons["head"]["left"].append(Polygon([
                            volume_points[8 * hd_ratio][j][k],
                            volume_points[8 * hd_ratio][j][k + 1],
                            volume_points[8 * hd_ratio][j + 1][k + 1],
                            volume_points[8 * hd_ratio][j + 1][k]],
                            color))

        if "top" in self.visible_faces["head"]["front"]:
            for i in range(0, 8 * hd_ratio):
                for k in range(-2 * hd_ratio, 6 * hd_ratio):
                    color = skin.getpixel((8 * hd_ratio + i, 2 * hd_ratio + k))
                    if color[3] != 0:
                        self.polygons["head"]["top"].append(Polygon([
                            volume_points[i][0][k],
                            volume_points[i + 1][0][k],
                            volume_points[i + 1][0][k + 1],
                            volume_points[i][0][k + 1]],
                            color))

        if "bottom" in self.visible_faces["head"]["front"]:
            for i in range(0, 8 * hd_ratio):
                for k in range(-2 * hd_ratio, 6 * hd_ratio):
                    color = skin.getpixel((16 * hd_ratio + i, 2 * hd_ratio + k))
                    if color[3] != 0:
                        self.polygons["head"]["bottom"].append(Polygon([
                            volume_points[i][8 * hd_ratio][k],
                            volume_points[i + 1][8 * hd_ratio][k],
                            volume_points[i + 1][8 * hd_ratio][k + 1],
                            volume_points[i][8 * hd_ratio][k + 1]],
                            color))

        """Helmet / Hair"""
        if self.display_hair:
            volume_points = {}
            for i in range(0, 9 * hd_ratio):
                for j in range(0, 9 * hd_ratio):
                    volume_points = append_dict(volume_points, i, j, -2 * hd_ratio,
                                                Point(self, {"x": i * 8.5 / 8 - 0.25 * hd_ratio,
                                                             "y": j * 8.5 / 8 - 0.25 * hd_ratio,
                                                             "z": -2.25 * hd_ratio}))
                    volume_points = append_dict(volume_points, i, j, 6 * hd_ratio,
                                                Point(self, {"x": i * 8.5 / 8 - 0.25 * hd_ratio,
                                                             "y": j * 8.5 / 8 - 0.25 * hd_ratio,
                                                             "z": 6.25 * hd_ratio}))

            for j in range(0, 9 * hd_ratio):
                for k in range(-2 * hd_ratio, 7 * hd_ratio):
                    volume_points = append_dict(volume_points, 0, j, k,
                                                Point(self, {"x": -0.25 * hd_ratio,
                                                             "y": j * 8.5 / 8 - 0.25 * hd_ratio,
                                                             "z": k * 8.5 / 8 - 0.25 * hd_ratio}))
                    volume_points = append_dict(volume_points, 8 * hd_ratio, j, k,
                                                Point(self, {"x": 8.25 * hd_ratio,
                                                             "y": j * 8.5 / 8 - 0.25 * hd_ratio,
                                                             "z": k * 8.5 / 8 - 0.25 * hd_ratio}))

            for i in range(0, 9 * hd_ratio):
                for k in range(-2 * hd_ratio, 7 * hd_ratio):
                    volume_points = append_dict(volume_points, i, 0, k,
                                                Point(self, {"x": i * 8.5 / 8 - 0.25 * hd_ratio,
                                                             "y": -0.25 * hd_ratio,
                                                             "z": k * 8.5 / 8 - 0.25 * hd_ratio}))
                    volume_points = append_dict(volume_points, i, 8 * hd_ratio, k,
                                                Point(self, {"x": i * 8.5 / 8 - 0.25 * hd_ratio,
                                                             "y": 8.25 * hd_ratio,
                                                             "z": k * 8.5 / 8 - 0.25 * hd_ratio}))

            for i in range(0, 8 * hd_ratio):
                for j in range(0, 8 * hd_ratio):
                    color = skin.getpixel((64 * hd_ratio - 1 - i, 8 * hd_ratio + j))
                    if color[3] != 0:
                        self.polygons["helmet"]["back"].append(Polygon([
                            volume_points[i][j][-2 * hd_ratio],
                            volume_points[i + 1][j][-2 * hd_ratio],
                            volume_points[i + 1][j + 1][-2 * hd_ratio],
                            volume_points[i][j + 1][-2 * hd_ratio]],
                            color))

            for i in range(0, 8 * hd_ratio):
                for j in range(0, 8 * hd_ratio):
                    color = skin.getpixel((40 * hd_ratio + i, 8 * hd_ratio + j))
                    if color[3] != 0:
                        self.polygons["helmet"]["front"].append(Polygon([
                            volume_points[i][j][6 * hd_ratio],
                            volume_points[i + 1][j][6 * hd_ratio],
                            volume_points[i + 1][j + 1][6 * hd_ratio],
                            volume_points[i][j + 1][6 * hd_ratio]],
                            color))

            for j in range(0, 8 * hd_ratio):
                for k in range(-2 * hd_ratio, 6 * hd_ratio):
                    color = skin.getpixel((34 * hd_ratio + k, 8 * hd_ratio + j))
                    if color[3] != 0:
                        self.polygons["helmet"]["right"].append(Polygon([
                            volume_points[0][j][k],
                            volume_points[0][j][k + 1],
                            volume_points[0][j + 1][k + 1],
                            volume_points[0][j + 1][k]],
                            color))

            for j in range(0, 8 * hd_ratio):
                for k in range(-2 * hd_ratio, 6 * hd_ratio):
                    color = skin.getpixel((54 * hd_ratio - k - 1, 8 * hd_ratio + j))
                    if color[3] != 0:
                        self.polygons["helmet"]["left"].append(Polygon([
                            volume_points[8 * hd_ratio][j][k],
                            volume_points[8 * hd_ratio][j][k + 1],
                            volume_points[8 * hd_ratio][j + 1][k + 1],
                            volume_points[8 * hd_ratio][j + 1][k]],
                            color))

            for i in range(0, 8 * hd_ratio):
                for k in range(-2 * hd_ratio, 6 * hd_ratio):
                    color = skin.getpixel((40 * hd_ratio + i, 2 * hd_ratio + k))
                    if color[3] != 0:
                        self.polygons["helmet"]["top"].append(Polygon([
                            volume_points[i][0][k],
                            volume_points[i + 1][0][k],
                            volume_points[i + 1][0][k + 1],
                            volume_points[i][0][k + 1]],
                            color))

            for i in range(0, 8 * hd_ratio):
                for k in range(-2 * hd_ratio, 6 * hd_ratio):
                    color = skin.getpixel((48 * hd_ratio + 1, 2 * hd_ratio + k))
                    if color[3] != 0:
                        self.polygons["helmet"]["bottom"].append(Polygon([
                            volume_points[i][8 * hd_ratio][k],
                            volume_points[i + 1][8 * hd_ratio][k],
                            volume_points[i + 1][8 * hd_ratio][k + 1],
                            volume_points[i][8 * hd_ratio][k + 1]],
                            color))

        if not self.head_only:
            """Torso"""
            volume_points = {}
            for i in range(0, 9 * hd_ratio):
                for j in range(0, 13 * hd_ratio):
                    volume_points = append_dict(volume_points, i, j, 0,
                                                Point(self, {"x": i, "y": j + 8 * hd_ratio, "z": 0}))
                    volume_points = append_dict(volume_points, i, j, 4 * hd_ratio,
                                                Point(self, {"x": i, "y": j + 8 * hd_ratio, "z": 4 * hd_ratio}))

            for j in range(0, 13 * hd_ratio):
                for k in range(0, 5 * hd_ratio):
                    volume_points = append_dict(volume_points, 0, j, k,
                                                Point(self, {"x": 0, "y": j + 8 * hd_ratio, "z": k}))
                    volume_points = append_dict(volume_points, 8 * hd_ratio, j, k,
                                                Point(self, {"x": 8 * hd_ratio, "y": j + 8 * hd_ratio, "z": k}))

            for i in range(0, 9 * hd_ratio):
                for k in range(0, 5 * hd_ratio):
                    volume_points = append_dict(volume_points, i, 0, k,
                                                Point(self, {"x": i, "y": 8 * hd_ratio, "z": k}))
                    volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                Point(self, {"x": i, "y": 20 * hd_ratio, "z": k}))

            if "back" in self.visible_faces["torso"]["front"]:
                for i in range(0, 8 * hd_ratio):
                    for j in range(0, 12 * hd_ratio):
                        color = skin.getpixel(((40 * hd_ratio - 1) - i, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["torso"]["back"].append(Polygon([
                                volume_points[i][j][0],
                                volume_points[i + 1][j][0],
                                volume_points[i + 1][j + 1][0],
                                volume_points[i][j + 1][0]],
                                color))

            if "front" in self.visible_faces["torso"]["front"]:
                for i in range(0, 8 * hd_ratio):
                    for j in range(0, 12 * hd_ratio):
                        color = skin.getpixel((20 * hd_ratio + i, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["torso"]["front"].append(Polygon([
                                volume_points[i][j][4 * hd_ratio],
                                volume_points[i + 1][j][4 * hd_ratio],
                                volume_points[i + 1][j + 1][4 * hd_ratio],
                                volume_points[i][j + 1][4 * hd_ratio]],
                                color))

            if "right" in self.visible_faces["torso"]["front"]:
                for j in range(0, 12 * hd_ratio):
                    for k in range(0 * hd_ratio, 4 * hd_ratio):
                        color = skin.getpixel((16 * hd_ratio + k, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["torso"]["right"].append(Polygon([
                                volume_points[0][j][k],
                                volume_points[0][j][k + 1],
                                volume_points[0][j + 1][k + 1],
                                volume_points[0][j + 1][k]],
                                color))

            if "left" in self.visible_faces["torso"]["front"]:
                for j in range(0, 12 * hd_ratio):
                    for k in range(0 * hd_ratio, 4 * hd_ratio):
                        color = skin.getpixel(((32 * hd_ratio - 1) - k, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["torso"]["left"].append(Polygon([
                                volume_points[8 * hd_ratio][j][k],
                                volume_points[8 * hd_ratio][j][k + 1],
                                volume_points[8 * hd_ratio][j + 1][k + 1],
                                volume_points[8 * hd_ratio][j + 1][k]],
                                color))

            if "top" in self.visible_faces["torso"]["front"]:
                for i in range(0, 8 * hd_ratio):
                    for k in range(0 * hd_ratio, 4 * hd_ratio):
                        color = skin.getpixel((20 * hd_ratio + i, 16 * hd_ratio + k))
                        if color[3] != 0:
                            self.polygons["torso"]["top"].append(Polygon([
                                volume_points[i][0][k],
                                volume_points[i + 1][0][k],
                                volume_points[i + 1][0][k + 1],
                                volume_points[i][0][k + 1]],
                                color))

            if "bottom" in self.visible_faces["torso"]["front"]:
                for i in range(0, 8 * hd_ratio):
                    for k in range(0 * hd_ratio, 4 * hd_ratio):
                        color = skin.getpixel((28 * hd_ratio + i, (20 * hd_ratio - 1) - k))
                        if color[3] != 0:
                            self.polygons["torso"]["bottom"].append(Polygon([
                                volume_points[i][12 * hd_ratio][k],
                                volume_points[i + 1][12 * hd_ratio][k],
                                volume_points[i + 1][12 * hd_ratio][k + 1],
                                volume_points[i][12 * hd_ratio][k + 1]],
                                color))

            """Torso 2nd layer"""
            if self.layers:
                volume_points = {}
                for i in range(0, 9 * hd_ratio):
                    for j in range(0, 13 * hd_ratio):
                        volume_points = append_dict(volume_points, i, j, 0,
                                                    Point(self, {"x": i * 8.25 / 8 - 0.125 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": -0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, i, j, 4 * hd_ratio,
                                                    Point(self, {"x": i * 8.25 / 8 - 0.125 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": 4.125 * hd_ratio}))

                for j in range(0, 13 * hd_ratio):
                    for k in range(0, 5 * hd_ratio):
                        volume_points = append_dict(volume_points, 0, j, k,
                                                    Point(self, {"x": -0.125 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, 8 * hd_ratio, j, k,
                                                    Point(self, {"x": 8.125 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))

                for i in range(0, 9 * hd_ratio):
                    for k in range(0, 5 * hd_ratio):
                        volume_points = append_dict(volume_points, i, 0, k,
                                                    Point(self, {"x": i * 8.25 / 8 - 0.125 * hd_ratio,
                                                                 "y": 7.875 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                    Point(self, {"x": i * 8.25 / 8 - 0.125 * hd_ratio,
                                                                 "y": 20.125 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))

                if "back" in self.visible_faces["torso_layer"]["front"]:
                    for i in range(0, 8 * hd_ratio):
                        for j in range(0, 12 * hd_ratio):
                            color = skin.getpixel(((40 * hd_ratio - 1) - i, 20 * hd_ratio + j + 16))
                            if color[3] != 0:
                                self.polygons["torso_layer"]["back"].append(Polygon([
                                    volume_points[i][j][0],
                                    volume_points[i + 1][j][0],
                                    volume_points[i + 1][j + 1][0],
                                    volume_points[i][j + 1][0]],
                                    color))

                if "front" in self.visible_faces["torso_layer"]["front"]:
                    for i in range(0, 8 * hd_ratio):
                        for j in range(0, 12 * hd_ratio):
                            color = skin.getpixel((20 * hd_ratio + i, 20 * hd_ratio + j + 16))
                            if color[3] != 0:
                                self.polygons["torso_layer"]["front"].append(Polygon([
                                    volume_points[i][j][4 * hd_ratio],
                                    volume_points[i + 1][j][4 * hd_ratio],
                                    volume_points[i + 1][j + 1][4 * hd_ratio],
                                    volume_points[i][j + 1][4 * hd_ratio]],
                                    color))

                if "right" in self.visible_faces["torso_layer"]["front"]:
                    for j in range(0, 12 * hd_ratio):
                        for k in range(0 * hd_ratio, 4 * hd_ratio):
                            color = skin.getpixel((16 * hd_ratio + k, 20 * hd_ratio + j + 16))
                            if color[3] != 0:
                                self.polygons["torso_layer"]["right"].append(Polygon([
                                    volume_points[0][j][k],
                                    volume_points[0][j][k + 1],
                                    volume_points[0][j + 1][k + 1],
                                    volume_points[0][j + 1][k]],
                                    color))

                if "left" in self.visible_faces["torso_layer"]["front"]:
                    for j in range(0, 12 * hd_ratio):
                        for k in range(0 * hd_ratio, 4 * hd_ratio):
                            color = skin.getpixel(((32 * hd_ratio - 1) - k, 20 * hd_ratio + j + 16))
                            if color[3] != 0:
                                self.polygons["torso_layer"]["left"].append(Polygon([
                                    volume_points[8 * hd_ratio][j][k],
                                    volume_points[8 * hd_ratio][j][k + 1],
                                    volume_points[8 * hd_ratio][j + 1][k + 1],
                                    volume_points[8 * hd_ratio][j + 1][k]],
                                    color))

                if "top" in self.visible_faces["torso_layer"]["front"]:
                    for i in range(0, 8 * hd_ratio):
                        for k in range(0 * hd_ratio, 4 * hd_ratio):
                            color = skin.getpixel((20 * hd_ratio + i, 16 * hd_ratio + k + 16))
                            if color[3] != 0:
                                self.polygons["torso_layer"]["top"].append(Polygon([
                                    volume_points[i][0][k],
                                    volume_points[i + 1][0][k],
                                    volume_points[i + 1][0][k + 1],
                                    volume_points[i][0][k + 1]],
                                    color))

                if "bottom" in self.visible_faces["torso_layer"]["front"]:
                    for i in range(0, 8 * hd_ratio):
                        for k in range(0 * hd_ratio, 4 * hd_ratio):
                            color = skin.getpixel((28 * hd_ratio + i, (20 * hd_ratio - 1) - k + 16))
                            if color[3] != 0:
                                self.polygons["torso_layer"]["bottom"].append(Polygon([
                                    volume_points[i][12 * hd_ratio][k],
                                    volume_points[i + 1][12 * hd_ratio][k],
                                    volume_points[i + 1][12 * hd_ratio][k + 1],
                                    volume_points[i][12 * hd_ratio][k + 1]],
                                    color))

            """Cape"""
            if self.display_cape:
                volume_points = {}
                for i in range(0, 11 * hd_ratio):
                    for j in range(0, 17 * hd_ratio):
                        volume_points = append_dict(volume_points, i, j, 0,
                                                    Point(self, {"x": i - 1, "y": j + 8 * hd_ratio, "z": -1}))
                        volume_points = append_dict(volume_points, i, j, 1 * hd_ratio,
                                                    Point(self, {"x": i - 1, "y": j + 8 * hd_ratio, "z": 0 * hd_ratio}))

                for j in range(0, 17 * hd_ratio):
                    for k in range(0, 2 * hd_ratio):
                        volume_points = append_dict(volume_points, 0, j, k,
                                                    Point(self, {"x": 0, "y": j + 8 * hd_ratio, "z": k}))
                        volume_points = append_dict(volume_points, 8 * hd_ratio, j, k,
                                                    Point(self, {"x": 8 * hd_ratio, "y": j + 8 * hd_ratio, "z": k}))

                for i in range(0, 11 * hd_ratio):
                    for k in range(0, 2 * hd_ratio):
                        volume_points = append_dict(volume_points, i, 0, k,
                                                    Point(self, {"x": i, "y": 8 * hd_ratio, "z": k}))
                        volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                    Point(self, {"x": i, "y": 20 * hd_ratio, "z": k}))

                if "back" in self.visible_faces["cape"]["front"]:
                    for i in range(0, 10 * hd_ratio):
                        for j in range(0, 16 * hd_ratio):
                            color = im_cape.getpixel(((11 * hd_ratio - 1) - i, 1 * hd_ratio + j))
                            if color[3] != 0:
                                self.polygons["cape"]["back"].append(Polygon([
                                    volume_points[i][j][0],
                                    volume_points[i + 1][j][0],
                                    volume_points[i + 1][j + 1][0],
                                    volume_points[i][j + 1][0]],
                                    color))

                if "front" in self.visible_faces["cape"]["front"]:
                    for i in range(0, 10 * hd_ratio):
                        for j in range(0, 16 * hd_ratio):
                            color = im_cape.getpixel((12 * hd_ratio + i, 1 * hd_ratio + j))
                            if color[3] != 0:
                                self.polygons["cape"]["front"].append(Polygon([
                                    volume_points[i][j][1 * hd_ratio],
                                    volume_points[i + 1][j][1 * hd_ratio],
                                    volume_points[i + 1][j + 1][1 * hd_ratio],
                                    volume_points[i][j + 1][1 * hd_ratio]],
                                    color))

                if "right" in self.visible_faces["cape"]["front"]:
                    for j in range(0, 16 * hd_ratio):
                        color = im_cape.getpixel((12 * hd_ratio, 1 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["cape"]["right"].append(Polygon([
                                volume_points[0][j][0],
                                volume_points[0][j][1],
                                volume_points[0][j + 1][1],
                                volume_points[0][j + 1][0]],
                                color))

                if "left" in self.visible_faces["cape"]["front"]:
                    for j in range(0, 16 * hd_ratio):
                        color = im_cape.getpixel((1 * hd_ratio, 1 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["cape"]["left"].append(Polygon([
                                volume_points[10 * hd_ratio][j][0],
                                volume_points[10 * hd_ratio][j][1],
                                volume_points[10 * hd_ratio][j + 1][1],
                                volume_points[10 * hd_ratio][j + 1][0]],
                                color))

                if "top" in self.visible_faces["cape"]["front"]:
                    for i in range(0, 10 * hd_ratio):
                        color = im_cape.getpixel((1 + i, 0))
                        if color[3] != 0:
                            self.polygons["cape"]["top"].append(Polygon([
                                volume_points[i][0][0],
                                volume_points[i + 1][0][0],
                                volume_points[i + 1][0][1],
                                volume_points[i][0][1]],
                                color))

                if "bottom" in self.visible_faces["cape"]["front"]:
                    for i in range(0, 10 * hd_ratio):
                        color = im_cape.getpixel((11 * hd_ratio + i, 0))
                        if color[3] != 0:
                            self.polygons["cape"]["bottom"].append(Polygon([
                                volume_points[i][16 * hd_ratio][0],
                                volume_points[i + 1][16 * hd_ratio][0],
                                volume_points[i + 1][16 * hd_ratio][1],
                                volume_points[i][16 * hd_ratio][1]],
                                color))

            start = 1 if self.player.is_slim else 0
            """Right arm"""
            volume_points = {}
            for i in range(start, 5 * hd_ratio):
                for j in range(0, 13 * hd_ratio):
                    volume_points = append_dict(volume_points, i, j, 0,
                                                Point(self, {"x": i - 4 * hd_ratio, "y": j + 8 * hd_ratio, "z": 0}))
                    volume_points = append_dict(volume_points, i, j, 4 * hd_ratio,
                                                Point(self, {"x": i - 4 * hd_ratio, "y": j + 8 * hd_ratio, "z": 4 * hd_ratio}))

            for j in range(0, 13 * hd_ratio):
                for k in range(0, 5 * hd_ratio):
                    volume_points = append_dict(volume_points, start, j, k,
                                                Point(self, {"x": -4 * hd_ratio + start, "y": j + 8 * hd_ratio, "z": k}))
                    volume_points = append_dict(volume_points, 4 * hd_ratio, j, k,
                                                Point(self, {"x": 0, "y": j + 8 * hd_ratio, "z": k}))

            for i in range(start, 5 * hd_ratio):
                for k in range(0, 5 * hd_ratio):
                    volume_points = append_dict(volume_points, i, 0, k,
                                                Point(self, {"x": i - 4 * hd_ratio, "y": 8 * hd_ratio, "z": k}))
                    volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                Point(self, {"x": i - 4 * hd_ratio, "y": 20 * hd_ratio, "z": k}))

            if "back" in self.visible_faces["r_arm"]["front"]:
                for i in range(start, 4 * hd_ratio):
                    for j in range(0, 12 * hd_ratio):
                        color = skin.getpixel((((56 - start) * hd_ratio - 1) - i, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["r_arm"]["back"].append(Polygon([
                                volume_points[i][j][0],
                                volume_points[i + 1][j][0],
                                volume_points[i + 1][j + 1][0],
                                volume_points[i][j + 1][0]],
                                color))

            if "front" in self.visible_faces["r_arm"]["front"]:
                for i in range(start, 4 * hd_ratio):
                    for j in range(0, 12 * hd_ratio):
                        color = skin.getpixel(((44 - start) * hd_ratio + i, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["r_arm"]["front"].append(Polygon([
                                volume_points[i][j][4 * hd_ratio],
                                volume_points[i + 1][j][4 * hd_ratio],
                                volume_points[i + 1][j + 1][4 * hd_ratio],
                                volume_points[i][j + 1][4 * hd_ratio]],
                                color))

            if "right" in self.visible_faces["r_arm"]["front"]:
                for j in range(0, 12 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((40 * hd_ratio + k, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["r_arm"]["right"].append(Polygon([
                                volume_points[start][j][k],
                                volume_points[start][j][k + 1],
                                volume_points[start][j + 1][k + 1],
                                volume_points[start][j + 1][k]],
                                color))

            if "left" in self.visible_faces["r_arm"]["front"]:
                for j in range(0, 12 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((((52 - start) * hd_ratio - 1) - k, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["r_arm"]["left"].append(Polygon([
                                volume_points[4 * hd_ratio][j][k],
                                volume_points[4 * hd_ratio][j][k + 1],
                                volume_points[4 * hd_ratio][j + 1][k + 1],
                                volume_points[4 * hd_ratio][j + 1][k]],
                                color))

            if "top" in self.visible_faces["r_arm"]["front"]:
                for i in range(start, 4 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel(((44 - start) * hd_ratio + i, 16 * hd_ratio + k))
                        if color[3] != 0:
                            self.polygons["r_arm"]["top"].append(Polygon([
                                volume_points[i][0][k],
                                volume_points[i + 1][0][k],
                                volume_points[i + 1][0][k + 1],
                                volume_points[i][0][k + 1]],
                                color))

            if "bottom" in self.visible_faces["r_arm"]["front"]:
                for i in range(start, 4 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel(((48 - start * 2) * hd_ratio + i, 16 * hd_ratio + k))
                        if color[3] != 0:
                            self.polygons["r_arm"]["bottom"].append(Polygon([
                                volume_points[i][12 * hd_ratio][k],
                                volume_points[i + 1][12 * hd_ratio][k],
                                volume_points[i + 1][12 * hd_ratio][k + 1],
                                volume_points[i][12 * hd_ratio][k + 1]],
                                color))

            """Right arm 2nd layer"""
            if self.layers:
                volume_points = {}
                for i in range(start, 5 * hd_ratio):
                    for j in range(0, 13 * hd_ratio):
                        volume_points = append_dict(volume_points, i, j, 0,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) - 4 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": -0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, i, j, 4 * hd_ratio,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) - 4 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": 4.125 * hd_ratio}))

                for j in range(0, 13 * hd_ratio):
                    for k in range(0, 5 * hd_ratio):
                        volume_points = append_dict(volume_points, start, j, k,
                                                    Point(self, {"x": (-4.125 + start) * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, 4 * hd_ratio, j, k,
                                                    Point(self, {"x": 0.125 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))

                for i in range(start, 5 * hd_ratio):
                    for k in range(0, 5 * hd_ratio):
                        volume_points = append_dict(volume_points, i, 0, k,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) - 4 * hd_ratio,
                                                                 "y": 7.875 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) - 4 * hd_ratio,
                                                                 "y": 20.125 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))

                if "back" in self.visible_faces["r_arm_layer"]["front"]:
                    for i in range(start, 4 * hd_ratio):
                        for j in range(0, 12 * hd_ratio):
                            color = skin.getpixel((((56 - start * 2) * hd_ratio - 1) - i, 20 * hd_ratio + j + 16))
                            if color[3] != 0:
                                self.polygons["r_arm_layer"]["back"].append(Polygon([
                                    volume_points[i][j][0],
                                    volume_points[i + 1][j][0],
                                    volume_points[i + 1][j + 1][0],
                                    volume_points[i][j + 1][0]],
                                    color))

                if "front" in self.visible_faces["r_arm_layer"]["front"]:
                    for i in range(start, 4 * hd_ratio):
                        for j in range(0, 12 * hd_ratio):
                            color = skin.getpixel(((44 - start) * hd_ratio + i, 20 * hd_ratio + j + 16))
                            if color[3] != 0:
                                self.polygons["r_arm_layer"]["front"].append(Polygon([
                                    volume_points[i][j][4 * hd_ratio],
                                    volume_points[i + 1][j][4 * hd_ratio],
                                    volume_points[i + 1][j + 1][4 * hd_ratio],
                                    volume_points[i][j + 1][4 * hd_ratio]],
                                    color))

                if "right" in self.visible_faces["r_arm_layer"]["front"]:
                    for j in range(0, 12 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel((40 * hd_ratio + k, 20 * hd_ratio + j + 16))
                            if color[3] != 0:
                                self.polygons["r_arm_layer"]["right"].append(Polygon([
                                    volume_points[start][j][k],
                                    volume_points[start][j][k + 1],
                                    volume_points[start][j + 1][k + 1],
                                    volume_points[start][j + 1][k]],
                                    color))

                if "left" in self.visible_faces["r_arm_layer"]["front"]:
                    for j in range(0, 12 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel((((52 - start) * hd_ratio - 1) - k, 20 * hd_ratio + j + 16))
                            if color[3] != 0:
                                self.polygons["r_arm_layer"]["left"].append(Polygon([
                                    volume_points[4 * hd_ratio][j][k],
                                    volume_points[4 * hd_ratio][j][k + 1],
                                    volume_points[4 * hd_ratio][j + 1][k + 1],
                                    volume_points[4 * hd_ratio][j + 1][k]],
                                    color))

                if "top" in self.visible_faces["r_arm_layer"]["front"]:
                    for i in range(start, 4 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel(((44 - start) * hd_ratio + i, 16 * hd_ratio + k + 16))
                            if color[3] != 0:
                                self.polygons["r_arm_layer"]["top"].append(Polygon([
                                    volume_points[i][0][k],
                                    volume_points[i + 1][0][k],
                                    volume_points[i + 1][0][k + 1],
                                    volume_points[i][0][k + 1]],
                                    color))

                if "bottom" in self.visible_faces["r_arm_layer"]["front"]:
                    for i in range(start, 4 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel(((48 - start * 2) * hd_ratio + i, 16 * hd_ratio + k + 16))
                            if color[3] != 0:
                                self.polygons["r_arm_layer"]["bottom"].append(Polygon([
                                    volume_points[i][12 * hd_ratio][k],
                                    volume_points[i + 1][12 * hd_ratio][k],
                                    volume_points[i + 1][12 * hd_ratio][k + 1],
                                    volume_points[i][12 * hd_ratio][k + 1]],
                                    color))

            """Left arm"""
            volume_points = {}
            for i in range(0, (5 - start) * hd_ratio):
                for j in range(0, 13 * hd_ratio):
                    volume_points = append_dict(volume_points, i, j, 0,
                                                Point(self, {"x": i + 8 * hd_ratio, "y": j + 8 * hd_ratio, "z": 0}))
                    volume_points = append_dict(volume_points, i, j, 4 * hd_ratio,
                                                Point(self, {"x": i + 8 * hd_ratio, "y": j + 8 * hd_ratio, "z": 4 * hd_ratio}))

            for j in range(0, 13 * hd_ratio):
                for k in range(0, 5 * hd_ratio):
                    volume_points = append_dict(volume_points, 0, j, k,
                                                Point(self, {"x": 8 * hd_ratio, "y": j + 8 * hd_ratio, "z": k}))
                    volume_points = append_dict(volume_points, (4 - start) * hd_ratio, j, k,
                                                Point(self, {"x": (12 - start) * hd_ratio, "y": j + 8 * hd_ratio, "z": k}))

            for i in range(0, (5 - start) * hd_ratio):
                for k in range(0, 5 * hd_ratio):
                    volume_points = append_dict(volume_points, i, 0, k,
                                                Point(self, {"x": i + 8 * hd_ratio, "y": 8 * hd_ratio, "z": k}))
                    volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                Point(self, {"x": i + 8 * hd_ratio, "y": 20 * hd_ratio, "z": k}))

            if "back" in self.visible_faces["l_arm"]["front"]:
                for i in range(0, (4 - start) * hd_ratio):
                    for j in range(0, 12 * hd_ratio):
                        color = skin.getpixel(((48 - start * 2) * hd_ratio - 1 - i, 52 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["l_arm"]["back"].append(Polygon([
                                volume_points[i][j][0],
                                volume_points[i + 1][j][0],
                                volume_points[i + 1][j + 1][0],
                                volume_points[i][j + 1][0]],
                                color))

            if "front" in self.visible_faces["l_arm"]["front"]:
                for i in range(0, (4 - start) * hd_ratio):
                    for j in range(0, 12 * hd_ratio):
                        color = skin.getpixel((36 * hd_ratio + i, 52 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["l_arm"]["front"].append(Polygon([
                                volume_points[i][j][4 * hd_ratio],
                                volume_points[i + 1][j][4 * hd_ratio],
                                volume_points[i + 1][j + 1][4 * hd_ratio],
                                volume_points[i][j + 1][4 * hd_ratio]],
                                color))

            if "right" in self.visible_faces["l_arm"]["front"]:
                for j in range(0, 12 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((32 * hd_ratio + k, 52 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["l_arm"]["right"].append(Polygon([
                                volume_points[0][j][k],
                                volume_points[0][j][k + 1],
                                volume_points[0][j + 1][k + 1],
                                volume_points[0][j + 1][k]],
                                color))

            if "left" in self.visible_faces["l_arm"]["front"]:
                for j in range(0, 12 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel(((44 - start) * hd_ratio - 1 - k, 52 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["l_arm"]["left"].append(Polygon([
                                volume_points[(4 - start) * hd_ratio][j][k],
                                volume_points[(4 - start) * hd_ratio][j][k + 1],
                                volume_points[(4 - start) * hd_ratio][j + 1][k + 1],
                                volume_points[(4 - start) * hd_ratio][j + 1][k]],
                                color))

            if "top" in self.visible_faces["l_arm"]["front"]:
                for i in range(0, (4 - start) * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((36 * hd_ratio + i, 48 * hd_ratio + k))
                        if color[3] != 0:
                            self.polygons["l_arm"]["top"].append(Polygon([
                                volume_points[i][0][k],
                                volume_points[i + 1][0][k],
                                volume_points[i + 1][0][k + 1],
                                volume_points[i][0][k + 1]],
                                color))

            if "bottom" in self.visible_faces["l_arm"]["front"]:
                for i in range(0, (4 - start) * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel(((40 - start) * hd_ratio + i, 48 * hd_ratio + k))
                        if color[3] != 0:
                            self.polygons["l_arm"]["bottom"].append(Polygon([
                                volume_points[i][12 * hd_ratio][k],
                                volume_points[i + 1][12 * hd_ratio][k],
                                volume_points[i + 1][12 * hd_ratio][k + 1],
                                volume_points[i][12 * hd_ratio][k + 1]],
                                color))

            """Left arm 2nd layer"""
            if self.layers:
                volume_points = {}
                for i in range(0, (5 - start) * hd_ratio):
                    for j in range(0, 13 * hd_ratio):
                        volume_points = append_dict(volume_points, i, j, 0,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": -0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, i, j, 4 * hd_ratio,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": 4.125 * hd_ratio}))

                for j in range(0, 13 * hd_ratio):
                    for k in range(0, 5 * hd_ratio):
                        volume_points = append_dict(volume_points, 0, j, k,
                                                    Point(self, {"x": 7.875 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, (4 - start) * hd_ratio, j, k,
                                                    Point(self, {"x": (12.125 - start) * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))

                for i in range(0, (5 - start) * hd_ratio):
                    for k in range(0, 5 * hd_ratio):
                        volume_points = append_dict(volume_points, i, 0, k,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "y": 7.875 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) + 8 * hd_ratio,
                                                                 "y": 20.125 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))

                if "back" in self.visible_faces["l_arm_layer"]["front"]:
                    for i in range(0, (4 - start) * hd_ratio):
                        for j in range(0, 12 * hd_ratio):
                            color = skin.getpixel(((64 - start * 2) * hd_ratio - 1 - i, 52 * hd_ratio + j))
                            if color[3] != 0:
                                self.polygons["l_arm_layer"]["back"].append(Polygon([
                                    volume_points[i][j][0],
                                    volume_points[i + 1][j][0],
                                    volume_points[i + 1][j + 1][0],
                                    volume_points[i][j + 1][0]],
                                    color))

                if "front" in self.visible_faces["l_arm_layer"]["front"]:
                    for i in range(0, (4 - start) * hd_ratio):
                        for j in range(0, 12 * hd_ratio):
                            color = skin.getpixel((52 * hd_ratio + i, 52 * hd_ratio + j))
                            if color[3] != 0:
                                self.polygons["l_arm_layer"]["front"].append(Polygon([
                                    volume_points[i][j][4 * hd_ratio],
                                    volume_points[i + 1][j][4 * hd_ratio],
                                    volume_points[i + 1][j + 1][4 * hd_ratio],
                                    volume_points[i][j + 1][4 * hd_ratio]],
                                    color))

                if "right" in self.visible_faces["l_arm_layer"]["front"]:
                    for j in range(0, 12 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel((48 * hd_ratio + k, 52 * hd_ratio + j))
                            if color[3] != 0:
                                self.polygons["l_arm_layer"]["right"].append(Polygon([
                                    volume_points[0][j][k],
                                    volume_points[0][j][k + 1],
                                    volume_points[0][j + 1][k + 1],
                                    volume_points[0][j + 1][k]],
                                    color))

                if "left" in self.visible_faces["l_arm_layer"]["front"]:
                    for j in range(0, 12 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel(((60 - start) * hd_ratio - 1 - k, 52 * hd_ratio + j))
                            if color[3] != 0:
                                self.polygons["l_arm_layer"]["left"].append(Polygon([
                                    volume_points[(4 - start) * hd_ratio][j][k],
                                    volume_points[(4 - start) * hd_ratio][j][k + 1],
                                    volume_points[(4 - start) * hd_ratio][j + 1][k + 1],
                                    volume_points[(4 - start) * hd_ratio][j + 1][k]],
                                    color))

                if "top" in self.visible_faces["l_arm_layer"]["front"]:
                    for i in range(0, (4 - start) * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel((52 * hd_ratio + i, 48 * hd_ratio + k))
                            if color[3] != 0:
                                self.polygons["l_arm_layer"]["top"].append(Polygon([
                                    volume_points[i][0][k],
                                    volume_points[i + 1][0][k],
                                    volume_points[i + 1][0][k + 1],
                                    volume_points[i][0][k + 1]],
                                    color))

                if "bottom" in self.visible_faces["l_arm_layer"]["front"]:
                    for i in range(0, (4 - start) * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel(((56 - start) * hd_ratio + i, 48 * hd_ratio + k))
                            if color[3] != 0:
                                self.polygons["l_arm_layer"]["bottom"].append(Polygon([
                                    volume_points[i][12 * hd_ratio][k],
                                    volume_points[i + 1][12 * hd_ratio][k],
                                    volume_points[i + 1][12 * hd_ratio][k + 1],
                                    volume_points[i][12 * hd_ratio][k + 1]],
                                    color))

            """Right leg"""
            volume_points = {}
            for i in range(0, 5 * hd_ratio):
                for j in range(0, 13 * hd_ratio):
                    volume_points = append_dict(volume_points, i, j, 0,
                                                Point(self, {"x": i, "y": j + 20 * hd_ratio, "z": 0}))
                    volume_points = append_dict(volume_points, i, j, 4 * hd_ratio,
                                                Point(self, {"x": i, "y": j + 20 * hd_ratio, "z": 4 * hd_ratio}))

            for j in range(0, 13 * hd_ratio):
                for k in range(0, 5 * hd_ratio):
                    volume_points = append_dict(volume_points, 0, j, k,
                                                Point(self, {"x": 0, "y": j + 20 * hd_ratio, "z": k}))
                    volume_points = append_dict(volume_points, 4 * hd_ratio, j, k,
                                                Point(self, {"x": 4 * hd_ratio, "y": j + 20 * hd_ratio, "z": k}))

            for i in range(0, 5 * hd_ratio):
                for k in range(0, 5 * hd_ratio):
                    volume_points = append_dict(volume_points, i, 0, k,
                                                Point(self, {"x": i, "y": 20 * hd_ratio, "z": k}))
                    volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                Point(self, {"x": i, "y": 32 * hd_ratio, "z": k}))

            if "back" in self.visible_faces["r_leg"]["front"]:
                for i in range(0, 4 * hd_ratio):
                    for j in range(0, 12 * hd_ratio):
                        color = skin.getpixel(((16 * hd_ratio - 1) - i, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["r_leg"]["back"].append(Polygon([
                                volume_points[i][j][0],
                                volume_points[i + 1][j][0],
                                volume_points[i + 1][j + 1][0],
                                volume_points[i][j + 1][0]],
                                color))

            if "front" in self.visible_faces["r_leg"]["front"]:
                for i in range(0, 4 * hd_ratio):
                    for j in range(0, 12 * hd_ratio):
                        color = skin.getpixel((4 * hd_ratio + i, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["r_leg"]["front"].append(Polygon([
                                volume_points[i][j][4 * hd_ratio],
                                volume_points[i + 1][j][4 * hd_ratio],
                                volume_points[i + 1][j + 1][4 * hd_ratio],
                                volume_points[i][j + 1][4 * hd_ratio]],
                                color))

            if "right" in self.visible_faces["r_leg"]["front"]:
                for j in range(0, 12 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((k, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["r_leg"]["right"].append(Polygon([
                                volume_points[0][j][k],
                                volume_points[0][j][k + 1],
                                volume_points[0][j + 1][k + 1],
                                volume_points[0][j + 1][k]],
                                color))

            if "left" in self.visible_faces["r_leg"]["front"]:
                for j in range(0, 12 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel(((12 * hd_ratio - 1) - k, 20 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["r_leg"]["left"].append(Polygon([
                                volume_points[4 * hd_ratio][j][k],
                                volume_points[4 * hd_ratio][j][k + 1],
                                volume_points[4 * hd_ratio][j + 1][k + 1],
                                volume_points[4 * hd_ratio][j + 1][k]],
                                color))

            if "top" in self.visible_faces["r_leg"]["front"]:
                for i in range(0, 4 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((4 * hd_ratio + i, 16 * hd_ratio + k))
                        if color[3] != 0:
                            self.polygons["r_leg"]["top"].append(Polygon([
                                volume_points[i][0][k],
                                volume_points[i + 1][0][k],
                                volume_points[i + 1][0][k + 1],
                                volume_points[i][0][k + 1]],
                                color))

            if "bottom" in self.visible_faces["r_leg"]["front"]:
                for i in range(0, 4 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((8 * hd_ratio + i, 16 * hd_ratio + k))
                        if color[3] != 0:
                            self.polygons["r_leg"]["bottom"].append(Polygon([
                                volume_points[i][12 * hd_ratio][k],
                                volume_points[i + 1][12 * hd_ratio][k],
                                volume_points[i + 1][12 * hd_ratio][k + 1],
                                volume_points[i][12 * hd_ratio][k + 1]],
                                color))

            """Right leg 2nd layer"""
            if self.layers:
                volume_points = {}
                for i in range(0, 5 * hd_ratio):
                    for j in range(0, 13 * hd_ratio):
                        volume_points = append_dict(volume_points, i, j, 0,
                                                    Point(self, {"x": i * 4.25 / 4 - 0.125 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 20 * hd_ratio,
                                                                 "z": -0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, i, j, 4 * hd_ratio,
                                                    Point(self, {"x": i * 4.25 / 4 - 0.125 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 20 * hd_ratio,
                                                                 "z": 4.125 * hd_ratio}))

                for j in range(0, 13 * hd_ratio):
                    for k in range(0, 5 * hd_ratio):
                        volume_points = append_dict(volume_points, 0, j, k,
                                                    Point(self, {"x": -0.125 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 20 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, 4 * hd_ratio, j, k,
                                                    Point(self, {"x": 4.125 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 20 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))

                for i in range(0, 5 * hd_ratio):
                    for k in range(0, 5 * hd_ratio):
                        volume_points = append_dict(volume_points, i, 0, k,
                                                    Point(self, {"x": i * 4.25 / 4 - 0.125 * hd_ratio,
                                                                 "y": 19.875 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                    Point(self, {"x": i * 4.25 / 4 - 0.125 * hd_ratio,
                                                                 "y": 32.125 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))

                if "back" in self.visible_faces["r_leg_layer"]["front"]:
                    for i in range(0, 4 * hd_ratio):
                        for j in range(0, 12 * hd_ratio):
                            color = skin.getpixel((16 * hd_ratio - 1 - i, 36 * hd_ratio + j))
                            if color[3] != 0:
                                self.polygons["r_leg_layer"]["back"].append(Polygon([
                                    volume_points[i][j][0],
                                    volume_points[i + 1][j][0],
                                    volume_points[i + 1][j + 1][0],
                                    volume_points[i][j + 1][0]],
                                    color))

                if "front" in self.visible_faces["r_leg_layer"]["front"]:
                    for i in range(0, 4 * hd_ratio):
                        for j in range(0, 12 * hd_ratio):
                            color = skin.getpixel((4 * hd_ratio + i, 36 * hd_ratio + j))
                            if color[3] != 0:
                                self.polygons["r_leg_layer"]["front"].append(Polygon([
                                    volume_points[i][j][4 * hd_ratio],
                                    volume_points[i + 1][j][4 * hd_ratio],
                                    volume_points[i + 1][j + 1][4 * hd_ratio],
                                    volume_points[i][j + 1][4 * hd_ratio]],
                                    color))

                if "right" in self.visible_faces["r_leg_layer"]["front"]:
                    for j in range(0, 12 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel((k, 36 * hd_ratio + j))
                            if color[3] != 0:
                                self.polygons["r_leg_layer"]["right"].append(Polygon([
                                    volume_points[0][j][k],
                                    volume_points[0][j][k + 1],
                                    volume_points[0][j + 1][k + 1],
                                    volume_points[0][j + 1][k]],
                                    color))

                if "left" in self.visible_faces["r_leg_layer"]["front"]:
                    for j in range(0, 12 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel((12 * hd_ratio - 1 - k, 36 * hd_ratio + j))
                            if color[3] != 0:
                                self.polygons["r_leg_layer"]["left"].append(Polygon([
                                    volume_points[4 * hd_ratio][j][k],
                                    volume_points[4 * hd_ratio][j][k + 1],
                                    volume_points[4 * hd_ratio][j + 1][k + 1],
                                    volume_points[4 * hd_ratio][j + 1][k]],
                                    color))

                if "top" in self.visible_faces["r_leg_layer"]["front"]:
                    for i in range(0, 4 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel((4 * hd_ratio + i, 32 * hd_ratio + k))
                            if color[3] != 0:
                                self.polygons["r_leg_layer"]["top"].append(Polygon([
                                    volume_points[i][0][k],
                                    volume_points[i + 1][0][k],
                                    volume_points[i + 1][0][k + 1],
                                    volume_points[i][0][k + 1]],
                                    color))

                if "bottom" in self.visible_faces["r_leg_layer"]["front"]:
                    for i in range(0, 4 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color = skin.getpixel((8 * hd_ratio + i, 32 * hd_ratio + k))
                            if color[3] != 0:
                                self.polygons["r_leg_layer"]["bottom"].append(Polygon([
                                    volume_points[i][12 * hd_ratio][k],
                                    volume_points[i + 1][12 * hd_ratio][k],
                                    volume_points[i + 1][12 * hd_ratio][k + 1],
                                    volume_points[i][12 * hd_ratio][k + 1]],
                                    color))

            """Left leg"""
            volume_points = {}
            for i in range(0, 9 * hd_ratio):
                for j in range(0, 13 * hd_ratio):
                    volume_points = append_dict(volume_points, i, j, 0,
                                                Point(self, {"x": i + 4 * hd_ratio,
                                                             "y": j + 20 * hd_ratio,
                                                             "z": 0}))
                    volume_points = append_dict(volume_points, i, j, 4 * hd_ratio,
                                                Point(self, {"x": i + 4 * hd_ratio,
                                                             "y": j + 20 * hd_ratio,
                                                             "z": 4 * hd_ratio}))

            for j in range(0, 13 * hd_ratio):
                for k in range(0, 5 * hd_ratio):
                    volume_points = append_dict(volume_points, 0, j, k,
                                                Point(self, {"x": 4 * hd_ratio, "y": j + 20 * hd_ratio, "z": k}))
                    volume_points = append_dict(volume_points, 4 * hd_ratio, j, k,
                                                Point(self, {"x": 8 * hd_ratio, "y": j + 20 * hd_ratio, "z": k}))

            for i in range(0, 9 * hd_ratio):
                for k in range(0, 5 * hd_ratio):
                    volume_points = append_dict(volume_points, i, 0, k,
                                                Point(self, {"x": i + 4 * hd_ratio, "y": 20 * hd_ratio, "z": k}))
                    volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                Point(self, {"x": i + 4 * hd_ratio, "y": 32 * hd_ratio, "z": k}))

            if "back" in self.visible_faces["l_leg"]["front"]:
                for i in range(0, 4 * hd_ratio):
                    for j in range(0, 12 * hd_ratio):
                        color = skin.getpixel((32 * hd_ratio - 1 - i, 52 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["l_leg"]["back"].append(Polygon([
                                volume_points[i][j][0],
                                volume_points[i + 1][j][0],
                                volume_points[i + 1][j + 1][0],
                                volume_points[i][j + 1][0]],
                                color))

            if "front" in self.visible_faces["l_leg"]["front"]:
                for i in range(0, 4 * hd_ratio):
                    for j in range(0, 12 * hd_ratio):
                        color = skin.getpixel((20 * hd_ratio + i, 52 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["l_leg"]["front"].append(Polygon([
                                volume_points[i][j][4 * hd_ratio],
                                volume_points[i + 1][j][4 * hd_ratio],
                                volume_points[i + 1][j + 1][4 * hd_ratio],
                                volume_points[i][j + 1][4 * hd_ratio]],
                                color))

            if "right" in self.visible_faces["l_leg"]["front"]:
                for j in range(0, 12 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((16 * hd_ratio + k, 52 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["l_leg"]["right"].append(Polygon([
                                volume_points[0][j][k],
                                volume_points[0][j][k + 1],
                                volume_points[0][j + 1][k + 1],
                                volume_points[0][j + 1][k]],
                                color))

            if "left" in self.visible_faces["l_leg"]["front"]:
                for j in range(0, 12 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((28 * hd_ratio - 1 - k, 52 * hd_ratio + j))
                        if color[3] != 0:
                            self.polygons["l_leg"]["left"].append(Polygon([
                                volume_points[4 * hd_ratio][j][k],
                                volume_points[4 * hd_ratio][j][k + 1],
                                volume_points[4 * hd_ratio][j + 1][k + 1],
                                volume_points[4 * hd_ratio][j + 1][k]],
                                color))

            if "top" in self.visible_faces["l_leg"]["front"]:
                for i in range(0, 4 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((20 * hd_ratio + i, 48 * hd_ratio + k))
                        if color[3] != 0:
                            self.polygons["l_leg"]["top"].append(Polygon([
                                volume_points[i][0][k],
                                volume_points[i + 1][0][k],
                                volume_points[i + 1][0][k + 1],
                                volume_points[i][0][k + 1]],
                                color))

            if "bottom" in self.visible_faces["l_leg"]["front"]:
                for i in range(0, 4 * hd_ratio):
                    for k in range(0, 4 * hd_ratio):
                        color = skin.getpixel((24 * hd_ratio + i, 48 * hd_ratio + k))
                        if color[3] != 0:
                            self.polygons["l_leg"]["bottom"].append(Polygon([
                                volume_points[i][12 * hd_ratio][k],
                                volume_points[i + 1][12 * hd_ratio][k],
                                volume_points[i + 1][12 * hd_ratio][k + 1],
                                volume_points[i][12 * hd_ratio][k + 1]],
                                color))

            """Left leg 2nd layer"""
            if self.layers:
                volume_points = {}
                for i in range(0, 5 * hd_ratio):
                    for j in range(0, 13 * hd_ratio):
                        volume_points = append_dict(volume_points, i, j, 0,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) + 4 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 20 * hd_ratio,
                                                                 "z": -0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, i, j, 4 * hd_ratio,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) + 4 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 20 * hd_ratio,
                                                                 "z": 4.125 * hd_ratio}))

                for j in range(0, 13 * hd_ratio):
                    for k in range(0, 5 * hd_ratio):
                        volume_points = append_dict(volume_points, 0, j, k,
                                                    Point(self, {"x": 3.875 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 20 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, 4 * hd_ratio, j, k,
                                                    Point(self, {"x": 8.125 * hd_ratio,
                                                                 "y": (j * 12.25 / 12 - 0.125 * hd_ratio) + 20 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))

                for i in range(0, 5 * hd_ratio):
                    for k in range(0, 5 * hd_ratio):
                        volume_points = append_dict(volume_points, i, 0, k,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) + 4 * hd_ratio,
                                                                 "y": 19.875 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))
                        volume_points = append_dict(volume_points, i, 12 * hd_ratio, k,
                                                    Point(self, {"x": (i * 4.25 / 4 - 0.125 * hd_ratio) + 4 * hd_ratio,
                                                                 "y": 32.125 * hd_ratio,
                                                                 "z": k * 4.25 / 4 - 0.125 * hd_ratio}))

                if "back" in self.visible_faces["l_leg_layer"]["front"]:
                    for i in range(0, 4 * hd_ratio):
                        for j in range(0, 12 * hd_ratio):
                            color1 = skin.getpixel((16 * hd_ratio - 1 - i, 52 * hd_ratio + j))
                            if color1[3] != 0:
                                self.polygons["l_leg_layer"]["back"].append(Polygon([
                                    volume_points[i][j][0],
                                    volume_points[i + 1][j][0],
                                    volume_points[i + 1][j + 1][0],
                                    volume_points[i][j + 1][0]],
                                    color1))

                if "front" in self.visible_faces["l_leg_layer"]["front"]:
                    for i in range(0, 4 * hd_ratio):
                        for j in range(0, 12 * hd_ratio):
                            color2 = skin.getpixel((4 * hd_ratio + i, 52 * hd_ratio + j))
                            if color2[3] != 0:
                                self.polygons["l_leg_layer"]["front"].append(Polygon([
                                    volume_points[i][j][4 * hd_ratio],
                                    volume_points[i + 1][j][4 * hd_ratio],
                                    volume_points[i + 1][j + 1][4 * hd_ratio],
                                    volume_points[i][j + 1][4 * hd_ratio]],
                                    color2))

                if "right" in self.visible_faces["l_leg_layer"]["front"]:
                    for j in range(0, 12 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color1 = skin.getpixel((k, 52 * hd_ratio + j))
                            if color1[3] != 0:
                                self.polygons["l_leg_layer"]["right"].append(Polygon([
                                    volume_points[0][j][k],
                                    volume_points[0][j][k + 1],
                                    volume_points[0][j + 1][k + 1],
                                    volume_points[0][j + 1][k]],
                                    color1))

                if "left" in self.visible_faces["l_leg_layer"]["front"]:
                    for j in range(0, 12 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color2 = skin.getpixel((12 * hd_ratio - 1 - k, 52 * hd_ratio + j))
                            if color2[3] != 0:
                                self.polygons["l_leg_layer"]["left"].append(Polygon([
                                    volume_points[4 * hd_ratio][j][k],
                                    volume_points[4 * hd_ratio][j][k + 1],
                                    volume_points[4 * hd_ratio][j + 1][k + 1],
                                    volume_points[4 * hd_ratio][j + 1][k]],
                                    color2))

                if "top" in self.visible_faces["l_leg_layer"]["front"]:
                    for i in range(0, 4 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color1 = skin.getpixel((4 * hd_ratio + i, 48 * hd_ratio + k))
                            if color1[3] != 0:
                                self.polygons["l_leg_layer"]["top"].append(Polygon([
                                    volume_points[i][0][k],
                                    volume_points[i + 1][0][k],
                                    volume_points[i + 1][0][k + 1],
                                    volume_points[i][0][k + 1]],
                                    color1))

                if "bottom" in self.visible_faces["l_leg_layer"]["front"]:
                    for i in range(0, 4 * hd_ratio):
                        for k in range(0, 4 * hd_ratio):
                            color2 = skin.getpixel((8 * hd_ratio + i, 48 * hd_ratio + k))
                            if color2[3] != 0:
                                self.polygons["l_leg_layer"]["bottom"].append(Polygon([
                                    volume_points[i][12 * hd_ratio][k],
                                    volume_points[i + 1][12 * hd_ratio][k],
                                    volume_points[i + 1][12 * hd_ratio][k + 1],
                                    volume_points[i][12 * hd_ratio][k + 1]],
                                    color2))

    def member_rotation(self, hd_ratio):
        for face in self.polygons["head"]:
            for poly in self.polygons["head"][face]:
                poly.pre_project(4 * hd_ratio, 8 * hd_ratio, 2 * hd_ratio, *self.body_angles["head"])

        if self.display_hair:
            for face in self.polygons["helmet"]:
                for poly in self.polygons["helmet"][face]:
                    poly.pre_project(4 * hd_ratio, 8 * hd_ratio, 2 * hd_ratio, *self.body_angles["head"])

        if not self.head_only:
            for face in self.polygons["cape"]:
                for poly in self.polygons["cape"][face]:
                    poly.pre_project(4 * hd_ratio, 8 * hd_ratio, 0 * hd_ratio, *self.body_angles["cape"])

            for face in self.polygons["r_arm"]:
                for poly in self.polygons["r_arm"][face]:
                    poly.pre_project(-2 * hd_ratio, 10 * hd_ratio, 2 * hd_ratio, *self.body_angles["r_arm"])

            for face in self.polygons["r_arm_layer"]:
                for poly in self.polygons["r_arm_layer"][face]:
                    poly.pre_project(-2 * hd_ratio, 10 * hd_ratio, 2 * hd_ratio, *self.body_angles["r_arm_layer"])

            for face in self.polygons["l_arm"]:
                for poly in self.polygons["l_arm"][face]:
                    poly.pre_project(10 * hd_ratio, 10 * hd_ratio, 2 * hd_ratio, *self.body_angles["l_arm"])

            for face in self.polygons["l_arm_layer"]:
                for poly in self.polygons["l_arm_layer"][face]:
                    poly.pre_project(10 * hd_ratio, 10 * hd_ratio, 2 * hd_ratio, *self.body_angles["l_arm_layer"])

            for face in self.polygons["r_leg"]:
                for poly in self.polygons["r_leg"][face]:
                    poly.pre_project(2 * hd_ratio, 22 * hd_ratio, 2 * hd_ratio, *self.body_angles["r_leg"])

            for face in self.polygons["r_leg_layer"]:
                for poly in self.polygons["r_leg_layer"][face]:
                    poly.pre_project(2 * hd_ratio, 22 * hd_ratio, 2 * hd_ratio, *self.body_angles["r_leg_layer"])

            for face in self.polygons["l_leg"]:
                for poly in self.polygons["l_leg"][face]:
                    poly.pre_project(6 * hd_ratio, 22 * hd_ratio, 2 * hd_ratio, *self.body_angles["l_leg"])

            for face in self.polygons["l_leg_layer"]:
                for poly in self.polygons["l_leg_layer"][face]:
                    poly.pre_project(6 * hd_ratio, 22 * hd_ratio, 2 * hd_ratio, *self.body_angles["l_leg_layer"])

    def create_project_plan(self):
        for piece in self.polygons:
            for face in self.polygons[piece]:
                for poly in self.polygons[piece][face]:
                    if not poly.is_projected:
                        poly.project()

    def display_image(self):
        width = self.max_x - self.min_x
        height = self.max_y - self.min_y
        ratio = self.ratio
        if ratio < 2:
            ratio = 2

        if self.aa:
            ratio *= 2

        src_width = ratio * width + 1
        src_height = ratio * height + 1
        real_width = src_width / 2
        real_height = src_height / 2

        image = Image.new('RGBA', (int(src_width), int(src_height)))

        display_order = self.get_display_order()
        draw = ImageDraw.Draw(image)
        for pieces in display_order:
            for piece, faces in pieces.items():
                for face in faces:
                    for poly in self.polygons[piece][face]:
                        poly.add_png_polygon(draw, self.min_x, self.min_y, ratio)

        if self.aa:
            image = image.resize((int(real_width), int(real_height)), resample=Image.LANCZOS)
        return image

    def get_display_order(self):
        display_order = []
        if "front" in self.front_faces:
            display_order.append({"cape": self.visible_faces["cape"]["front"]})

        if "top" in self.front_faces:
            if "right" in self.front_faces:
                display_order.append({"l_leg_layer": self.back_faces})
                display_order.append({"l_leg": self.visible_faces["l_leg"]["front"]})
                display_order.append({"l_leg_layer": self.visible_faces["l_leg"]["front"]})

                display_order.append({"r_leg_layer": self.back_faces})
                display_order.append({"r_leg": self.visible_faces["r_leg"]["front"]})
                display_order.append({"r_leg_layer": self.visible_faces["r_leg"]["front"]})

                display_order.append({"l_arm_layer": self.back_faces})
                display_order.append({"l_arm": self.visible_faces["l_arm"]["front"]})
                display_order.append({"l_arm_layer": self.visible_faces["l_arm"]["front"]})

                display_order.append({"torso_layer": self.back_faces})
                display_order.append({"torso": self.visible_faces["torso"]["front"]})
                display_order.append({"torso_layer": self.visible_faces["torso"]["front"]})

                display_order.append({"r_arm_layer": self.back_faces})
                display_order.append({"r_arm": self.visible_faces["r_arm"]["front"]})
                display_order.append({"r_arm_layer": self.visible_faces["r_arm"]["front"]})
            else:
                display_order.append({"r_leg_layer": self.back_faces})
                display_order.append({"r_leg": self.visible_faces["r_leg"]["front"]})
                display_order.append({"r_leg_layer": self.visible_faces["r_leg"]["front"]})

                display_order.append({"l_leg_layer": self.back_faces})
                display_order.append({"l_leg": self.visible_faces["l_leg"]["front"]})
                display_order.append({"l_leg_layer": self.visible_faces["l_leg"]["front"]})

                display_order.append({"r_arm_layer": self.back_faces})
                display_order.append({"r_arm": self.visible_faces["r_arm"]["front"]})
                display_order.append({"r_arm_layer": self.visible_faces["r_arm"]["front"]})

                display_order.append({"torso_layer": self.back_faces})
                display_order.append({"torso": self.visible_faces["torso"]["front"]})
                display_order.append({"torso_layer": self.visible_faces["torso"]["front"]})

                display_order.append({"l_arm_layer": self.back_faces})
                display_order.append({"l_arm": self.visible_faces["l_arm"]["front"]})
                display_order.append({"l_arm_layer": self.visible_faces["l_arm"]["front"]})

            if "back" in self.front_faces:
                display_order.append({"cape": self.visible_faces["cape"]["front"]})

            display_order.append({"helmet": self.back_faces})
            display_order.append({"head": self.visible_faces["head"]["front"]})
            display_order.append({"helmet": self.visible_faces["head"]["front"]})
        else:
            display_order.append({"helmet": self.back_faces})
            display_order.append({"head": self.visible_faces["head"]["front"]})
            display_order.append({"helmet": self.visible_faces["head"]["front"]})
            if "right" in self.front_faces:
                display_order.append({"l_arm_layer": self.back_faces})
                display_order.append({"l_arm": self.visible_faces["l_arm"]["front"]})
                display_order.append({"l_arm_layer": self.visible_faces["l_arm"]["front"]})

                display_order.append({"torso_layer": self.back_faces})
                display_order.append({"torso": self.visible_faces["torso"]["front"]})
                display_order.append({"torso_layer": self.visible_faces["torso"]["front"]})

                display_order.append({"r_arm_layer": self.back_faces})
                display_order.append({"r_arm": self.visible_faces["r_arm"]["front"]})
                display_order.append({"r_arm_layer": self.visible_faces["r_arm"]["front"]})

                display_order.append({"l_leg_layer": self.back_faces})
                display_order.append({"l_leg": self.visible_faces["l_leg"]["front"]})
                display_order.append({"l_leg_layer": self.visible_faces["l_leg"]["front"]})

                display_order.append({"r_leg_layer": self.back_faces})
                display_order.append({"r_leg": self.visible_faces["r_leg"]["front"]})
                display_order.append({"r_leg_layer": self.visible_faces["r_leg"]["front"]})
            else:
                display_order.append({"r_arm_layer": self.back_faces})
                display_order.append({"r_arm": self.visible_faces["r_arm"]["front"]})
                display_order.append({"r_arm_layer": self.visible_faces["r_arm"]["front"]})

                display_order.append({"torso_layer": self.back_faces})
                display_order.append({"torso": self.visible_faces["torso"]["front"]})
                display_order.append({"torso_layer": self.visible_faces["torso"]["front"]})

                display_order.append({"l_arm_layer": self.back_faces})
                display_order.append({"l_arm": self.visible_faces["l_arm"]["front"]})
                display_order.append({"l_arm_layer": self.visible_faces["l_arm"]["front"]})

                display_order.append({"r_leg_layer": self.back_faces})
                display_order.append({"r_leg": self.visible_faces["r_leg"]["front"]})
                display_order.append({"r_leg_layer": self.visible_faces["r_leg"]["front"]})

                display_order.append({"l_leg_layer": self.back_faces})
                display_order.append({"l_leg": self.visible_faces["l_leg"]["front"]})
                display_order.append({"l_leg_layer": self.visible_faces["l_leg"]["front"]})

            if "back" in self.front_faces:
                display_order.append({"cape": self.visible_faces["cape"]["front"]})

        return display_order


class Point:
    def __init__(self, super_cls, origin_coords, dest_coords={}, is_projected=False, is_pre_projected=False):
        self.dest_coords = dest_coords
        self.is_projected = is_projected
        self.is_pre_projected = is_pre_projected
        self.super = super_cls

        if (isinstance(origin_coords, dict)) and (len(origin_coords.keys()) == 3):
            x = origin_coords["x"] if not is_not_existing(origin_coords, "x") else 0
            y = origin_coords["y"] if not is_not_existing(origin_coords, "y") else 0
            z = origin_coords["z"] if not is_not_existing(origin_coords, "z") else 0
        else:
            x, y, z = 0, 0, 0

        self.origin_coords = {"x": x, "y": y, "z": z}

    def pre_project(self, dx, dy, dz, cos_a, sin_a, cos_b, sin_b):
        if not self.is_pre_projected:
            x = self.origin_coords["x"] - dx
            y = self.origin_coords["y"] - dy
            z = self.origin_coords["z"] - dz
            self.origin_coords["x"] = x * cos_b + z * sin_b + dx
            self.origin_coords["y"] = x * sin_a * sin_b + y * cos_a - z * sin_a * cos_b + dy
            self.origin_coords["z"] = -x * cos_a * sin_b + y * sin_a + z * cos_a * cos_b + dz
            self.is_pre_projected = True

    def project(self):
        x = self.origin_coords["x"]
        y = self.origin_coords["y"]
        z = self.origin_coords["z"]
        self.dest_coords = {}
        self.dest_coords["x"] = x * self.super.cos_b + z * self.super.sin_b
        self.dest_coords[
            "y"] = x * self.super.sin_a * self.super.sin_b + y * self.super.cos_a - z * self.super.sin_a * self.super.cos_b
        self.dest_coords[
            "z"] = -x * self.super.cos_a * self.super.sin_b + y * self.super.sin_a + z * self.super.cos_a * self.super.cos_b
        self.is_projected = True
        self.super.min_x = min(self.super.min_x, self.dest_coords["x"])
        self.super.max_x = max(self.super.max_x, self.dest_coords["x"])
        self.super.min_y = min(self.super.min_y, self.dest_coords["y"])
        self.super.max_y = max(self.super.max_y, self.dest_coords["y"])

    def get_depth(self):
        if not self.is_projected:
            self.project()
        return self.dest_coords["z"]

    def get_dest_coords(self):
        return self.dest_coords


class Polygon:
    def __init__(self, dots, color, is_projected=False, face="w", face_depth=0):
        self.face = face
        self.is_projected = is_projected
        self.face_depth = face_depth
        self.dots = dots
        self.color = color
        coord_0 = dots[0].origin_coords
        coord_1 = dots[1].origin_coords
        coord_2 = dots[2].origin_coords

        if (coord_0["x"] == coord_1["x"]) and (coord_1["x"] == coord_2["x"]):
            self.face = "x"
            self.face_depth = coord_0["x"]
        elif (coord_0["y"] == coord_1["y"]) and (coord_1["y"] == coord_2["y"]):
            self.face = "y"
            self.face_depth = coord_0["y"]
        elif (coord_0["z"] == coord_1["z"]) and (coord_1["z"] == coord_2["z"]):
            self.face = "z"
            self.face_depth = coord_0["z"]

    def add_png_polygon(self, draw, min_x, min_y, ratio):
        points_2d = []

        same_plan_x = True
        same_plan_y = True
        coord_x = None
        coord_y = None

        for dot in self.dots:
            coord = dot.dest_coords
            if not coord_x:
                coord_x = coord["x"]
            if not coord_y:
                coord_y = coord["y"]
            if coord_x != coord["x"]:
                same_plan_x = False
            if coord_y != coord["y"]:
                same_plan_y = False
            points_2d.append(((coord["x"] - min_x) * ratio, (coord["y"] - min_y) * ratio))

            if not (same_plan_x or same_plan_y):
                draw.polygon(points_2d, fill=self.color, outline=self.color)

    def project(self):
        for dot in self.dots:
            if not dot.is_projected:
                dot.project()
        self.is_projected = True

    def pre_project(self, dx, dy, dz, cos_a, sin_a, cos_b, sin_b):
        for dot in self.dots:
            dot.pre_project(dx, dy, dz, cos_a, sin_a, cos_b, sin_b)



import base64
import typing
import aiohttp
import json

from PIL import Image
from io import BytesIO
from typing import Optional

from .skin import Skin

if typing.TYPE_CHECKING:
    from .player import Player


__all__ = [
    "uuid_to_dashed",
    "uuid_to_undashed",
    "name_to_uuid",
    "uuid_to_name",
    "fetch_skin",
    "fetch_optifine_cape",
    "fetch_labymod_cape",
    "fetch_5zig_cape",
    "fetch_minecraftcapes_cape",
    "fetch_tlauncher_cape",
    "get_players_by_name",
]


def uuid_to_dashed(uuid: str) -> str:
    """Converts a not dashed UUID to a dashed one

    Parameters
    ----------
    uuid: str
        The UUID to convert

    Returns
    -------
    str
        The converted UUID
    """
    return f"{uuid[:8]}-{uuid[8:12]}-{uuid[12:16]}-{uuid[16:20]}-{uuid[20:]}"


def uuid_to_undashed(uuid: str) -> str:
    """Converts a dashed UUID to a not dashed one

    Parameters
    ----------
    uuid: str
        The UUID to convert

    Returns
    -------
    str
        The converted UUID
    """
    return uuid.replace("-", "")


async def name_to_uuid(name: str, session: aiohttp.ClientSession = None) -> Optional[str]:
    """Convert a minecraft name to a UUID

    Parameters
    ----------
    name: str
        The minecraft name to get the UUID for
    session: aiohttp.ClientSession
        The ClientSession to use for requests
        Defaults to a new session which is closed again after handling all requests

    Returns
    -------
    Optional[str]
        None if the given name is invalid
    """
    if session is None:
        session = aiohttp.ClientSession()
        close = True
    else:
        close = False

    async with session.get(f"https://api.mojang.com/users/profiles/minecraft/{name}") as resp:
        if resp.status == 200:
            uuid = (await resp.json())["id"]
        else:
            uuid = None

    if close:
        await session.close()

    return uuid


async def uuid_to_name(uuid: str, session: aiohttp.ClientSession = None) -> Optional[str]:
    """Convert a UUID to a minecraft name

    Parameters
    ----------
    uuid: str
        The UUID to get the minecraft name for
    session: aiohttp.ClientSession
        The ClientSession to use for requests
        Defaults to a new session which is closed again after handling all requests

    Returns
    -------
    Optional[str]
        None if the given UUID is invalid
    """
    if session is None:
        session = aiohttp.ClientSession()
        close = True
    else:
        close = False

    async with session.get(
            f"https://sessionserver.mojang.com/session/minecraft/profile/{uuid}"
    ) as resp:
        if resp.status == 200:
            name = (await resp.json())["name"]
        else:
            name = None

    if close:
        await session.close()

    return name


async def fetch_skin(
        player: "Player" = None,
        name: str = None,
        uuid: str = None,
        session: aiohttp.ClientSession = None
) -> Optional[Skin]:
    """Fetch a players skin

    Note
    ----
    This function also returns the players mojang cape (if available).

    Tip
    ----
    Passing a :py:class:`minepi.Player` is recommended. If none is available, a name or UUID can be passed.
    UUIDs are preferred over names in this case since they require one API call less.

    Parameters
    ----------
    player: Player
        The player to fetch the skin for
    name: str
        Minecraft username to fetch the skin for
    uuid: str
        UUID to fetch the skin for
    session: aiohttp.ClientSession
        The ClientSession to use for requests
        Defaults to a new session which is closed again after handling all requests

    Returns
    -------
    Skin
        A fully functional :py:class:`minepi.Skin` class

    Raises
    ------
    ValueError
        No :py:class:`minepi.Player`, name or UUID has been passed
    """
    if player is None and name is None and uuid is None:
        raise ValueError("At least one parameter must be passed")

    if session is None:
        session = aiohttp.ClientSession()
        close = True
    else:
        close = False

    if player and player.uuid is not None:
        uuid = player.uuid

    if uuid is None and name is not None:
        uuid = await name_to_uuid(name, session)

    if uuid is not None:
        async with session.get(
            f"https://sessionserver.mojang.com/session/minecraft/profile/{uuid}"
        ) as resp:
            if resp.status == 200:
                cape = None
                skin = None
                resp_dict = await resp.json()
                for p in resp_dict["properties"]:
                    if p["name"] == "textures":
                        textures = json.loads(base64.b64decode(p["value"]))["textures"]
                        skin_url = textures["SKIN"]["url"]
                        cape_url = textures["CAPE"]["url"] if "CAPE" in textures.keys() else None

                        if skin_url:
                            async with session.get(skin_url) as resp_skin:
                                if resp.status == 200:
                                    skin = Image.open(BytesIO(await resp_skin.read()))

                        if cape_url:
                            async with session.get(cape_url) as resp_cape:
                                if resp.status == 200:
                                    cape = Image.open(BytesIO(await resp_cape.read()))
                        break

    if close is True:
        await session.close()

    if skin is None:
        raise ValueError

    return Skin(
        raw_skin=skin,
        raw_cape=cape,
        raw_skin_url=skin_url,
        raw_cape_url=cape_url,
        name=resp_dict["name"]
    )


async def fetch_mojang_cape(
        player: "Player" = None,
        name: str = None,
        uuid: str = None,
        session: aiohttp.ClientSession = None
) -> Optional[Image.Image]:
    """Fetch a players mojang cape

    Tip
    ----
    Passing a :py:class:`minepi.Player` object is recommended. If none is available, a name or UUID can be passed.
    UUIDs are preferred over names in this case since they require one API call less.

    Parameters
    ----------
    player: Player
        The player to fetch the mojang cape for
    name: str
        Minecraft username to fetch the mojang cape for
    uuid: str
        UUID to fetch the mojang cape for
    session: aiohttp.ClientSession
        The ClientSession to use for requests
        Defaults to a new session which is closed again after handling all requests

    Returns
    -------
    Optional[PIL.Image.Image]
        None if the given player has no mojang cape

    Raises
    ------
    ValueError
        No :py:class:`minepi.Player`, name or UUID has been passed
    """
    s = await fetch_skin(player=player, name=name, uuid=uuid, session=session)
    return s.raw_cape if s is not None else None


async def fetch_optifine_cape(
        player: "Player" = None,
        name: str = None,
        uuid: str = None,
        session: aiohttp.ClientSession = None
) -> Optional[Image.Image]:
    """Fetch a players optifine cape

    Tip
    ----
    Passing a :py:class:`minepi.Player` object is recommended. If none is available, a name or UUID can be passed.
    Names are preferred over UUIDs in this case since they require one API call less.

    Parameters
    ----------
    player: Player
        The player to fetch the optifine cape for
    name: str
        Minecraft username to fetch the optifine cape for
    uuid: str
        UUID to fetch the optifine cape for
    session: aiohttp.ClientSession
        The ClientSession to use for requests
        Defaults to a new session which is closed again after handling all requests

    Returns
    -------
    Optional[PIL.Image.Image]
        None if the given player has no optifine cape

    Raises
    ------
    ValueError
        No :py:class:`minepi.Player`, name or UUID has been passed
    """
    if player is None and name is None and uuid is None:
        raise ValueError("At least one parameter must be passed")

    if session is None:
        session = aiohttp.ClientSession()
        close = True
    else:
        close = False

    if player is not None:
        name = player.name
    elif name is not None:
        pass
    else:
        name = await uuid_to_name(uuid, session)

    if name is not None:
        async with session.get(f"http://s.optifine.net/capes/{name}.png") as resp:
            if resp.status == 200:
                cape = Image.open(BytesIO(await resp.read()))
            else:
                cape = None
    else:
        cape = None

    if close:
        await session.close()

    return cape


async def fetch_labymod_cape(
        player: "Player" = None,
        name: str = None,
        uuid: str = None,
        session: aiohttp.ClientSession = None
) -> Optional[Image.Image]:
    """Fetch a players labymod cape

    Tip
    ----
    Passing a :py:class:`minepi.Player` object is recommended. If none is available, a name or UUID can be passed.
    UUIDs are preferred over names in this case since they require one API call less.

    Parameters
    ----------
    player: Player
        The player to fetch the labymod cape for
    name: str
        Minecraft username to fetch the labymod cape for
    uuid: str
        UUID to fetch the labymod cape for
    session: aiohttp.ClientSession
        The ClientSession to use for requests
        Defaults to a new session which is closed again after handling all requests

    Returns
    -------
    Optional[PIL.Image.Image]
        None if the given player has no labymod cape

    Raises
    ------
    ValueError
        No :py:class:`minepi.Player`, name or UUID has been passed
    """
    if player is None and name is None and uuid is None:
        raise ValueError("At least one parameter must be passed")

    if session is None:
        session = aiohttp.ClientSession()
        close = True
    else:
        close = False

    if player is not None:
        uuid = player.uuid
    elif uuid is not None:
        pass
    else:
        uuid = await name_to_uuid(name, session)

    if uuid is not None:
        if len(uuid) == 32:
            uuid = uuid_to_dashed(uuid)

        async with session.get(f"https://dl.labymod.net/capes/{uuid}") as resp:
            if resp.status == 200:
                cape = Image.open(BytesIO(await resp.read()))
            else:
                cape = None
    else:
        cape = None

    if close:
        await session.close()

    return cape


async def fetch_5zig_cape(
        player: "Player" = None,
        name: str = None,
        uuid: str = None,
        session: aiohttp.ClientSession = None
) -> Optional[Image.Image]:
    """Fetch a players 5Zig cape

    Tip
    ----
    Passing a :class:`minepi.Player` object is recommended. If none is available, a name or UUID can be passed.
    UUIDs are preferred over names in this case since they require one API call less.

    Warning
    -------
    5zig capes are not guaranteed to work yet.
    If you own an account with 5zigreborn cape, it would be very helpful if you could send us the
    username.

    Parameters
    ----------
    player: Player
        The player to fetch the 5Zig cape for
    name: str
        Minecraft username to fetch the 5Zig cape for
    uuid: str
        UUID to fetch the 5Zig cape for
    session: aiohttp.ClientSession
        The ClientSession to use for requests
        Defaults to a new session which is closed again after handling all requests

    Returns
    -------
    Optional[PIL.Image.Image]
        None if the given player has no 5Zig cape

    Raises
    ------
    ValueError
        No :py:class:`minepi.Player`, name or UUID has been passed
    """
    if player is None and name is None and uuid is None:
        raise ValueError("At least one parameter must be passed")

    if session is None:
        session = aiohttp.ClientSession()
        close = True
    else:
        close = False

    if player is not None:
        uuid = player.uuid
    elif uuid is not None:
        pass
    else:
        uuid = await name_to_uuid(name, session)

    if uuid is not None:
        if len(uuid) == 32:
            uuid = uuid_to_dashed(uuid)

        async with session.get(f"https://textures.5zigreborn.eu/profile/{uuid}") as resp:
            if resp.status == 200:
                cape = Image.open(BytesIO(await resp.read()))
            else:
                cape = None
    else:
        cape = None

    if close:
        await session.close()

    return cape


async def fetch_minecraftcapes_cape(
        player: "Player" = None,
        name: str = None,
        uuid: str = None,
        session: aiohttp.ClientSession = None
) -> Optional[Image.Image]:
    """Fetch a players MinecraftCapes cape

    Tip
    ----
    Passing a :py:class:`minepi.Player` object is recommended. If none is available, a name or UUID can be passed.
    UUIDs are preferred over names in this case since they require one API call less.

    Warning
    -------
    Animated MinecraftCapes capes are not supported yet

    Parameters
    ----------
    player: Player
        The player to fetch the MinecraftCapes cape for
    name: str
        Minecraft username to fetch the MinecraftCapes cape for
    uuid: str
        UUID to fetch the MinecraftCapes cape for
    session: aiohttp.ClientSession
        The ClientSession to use for requests
        Defaults to a new session which is closed again after handling all requests

    Returns
    -------
    Optional[PIL.Image.Image]
        None if the given player has no MinecraftCapes cape

    Raises
    ------
    ValueError
        No :py:class:`minepi.Player`, name or UUID has been passed
    """
    if player is None and name is None and uuid is None:
        raise ValueError("At least one parameter must be passed")

    if not session:
        session = aiohttp.ClientSession()
        close = True
    else:
        close = False

    if player is not None:
        uuid = player.uuid
    elif uuid is not None:
        pass
    else:
        uuid = await name_to_uuid(name, session)

    if uuid is not None:
        if len(uuid) == 36:
            uuid = uuid_to_undashed(uuid)

        async with session.get(f"https://minecraftcapes.net/profile/{uuid}/cape") as resp:
            if resp.status == 200:
                cape = Image.open(BytesIO(await resp.read()))
            else:
                cape = None
    else:
        cape = None

    if close:
        await session.close()

    return cape


async def fetch_tlauncher_cape(
        player: "Player" = None,
        name: str = None,
        uuid: str = None,
        session: aiohttp.ClientSession = None
) -> Optional[Image.Image]:
    """Fetch a players TLauncher cape

    Tip
    ----
    Passing a :py:class:`minepi.Player` object is recommended. If none is available, a name or UUID can be passed.
    Names are preferred over UUIDs in this case since they require one API call less.


    Warning
    -------
    TLauncher capes are still experimental and not guaranteed to work.
    If you want to help implement them further, please get in touch with me.
    If you could share a few account names with TLauncher capes (especially animated ones),
    that would help us a lot.

    Parameters
    ----------
    player: Player
        The player to fetch the TLauncher cape for
    name: str
        Minecraft username to fetch the TLauncher cape for
    uuid: str
        UUID to fetch the TLauncher cape for
    session: aiohttp.ClientSession
        The ClientSession to use for requests
        Defaults to a new session which is closed again after handling all requests

    Returns
    -------
    Optional[PIL.Image.Image]
        None if the given player has no TLauncher cape

    Raises
    ------
    ValueError
        No :py:class:`minepi.Player`, name or UUID has been passed
    """
    if player is None and name is None and uuid is None:
        raise ValueError()

    if session is None:
        session = aiohttp.ClientSession()
        close = True
    else:
        close = False

    if player is not None:
        name = player.name
    elif name is not None:
        pass
    else:
        name = await uuid_to_name(uuid, session)

    cape = None
    if name is not None:
        async with session.get(f"https://auth.tlauncher.org/skin/profile/texture/login/{name}") as resp:
            if resp.status == 200:
                resp_dict = await resp.json()
                cape_url = resp_dict["CAPE"]["url"] if "CAPE" in resp_dict.keys() else None
            else:
                cape_url = None

        if cape_url is not None:
            if not cape_url.startswith("http://textures.minecraft.net/"):
                async with session.get(cape_url) as resp:
                    if resp.status == 200:
                        cape = Image.open(BytesIO(await resp.read()))

    if close:
        await session.close()

    return cape


async def get_players_by_name(names: list, session: aiohttp.ClientSession = None):
    """Useful helper function to get multiple :py:class:`minepi.Player` objects

    Only does one API call for the entire list instead of one per player
    This is recommended to be used if you have a list of usernames

    Parameters
    ----------
    names: list
        A list of minecraft usernames
    session: aiohttp.ClientSession
        The ClientSession to use for requests
        Defaults to a new session which is closed again after handling all requests

    Returns
    -------
    list
        A list of :py:class:`minepi.Player` objects
    """
    if session is None:
        session = aiohttp.ClientSession()
        close = True
    else:
        close = False

    players = []
    async with session.post("https://api.mojang.com/profiles/minecraft", json=names) as resp:
        if resp.status == 200:
            for entry in await resp.json():
                players.append(Player(uuid=entry["id"], name=entry["name"], session=session if not close else None))

    if close:
        await session.close()

    return players
